# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import argparse
import warnings
import traceback

from .message import MessageCenter


def _add_gen_annotation_parser(subparsers):
    parser = subparsers.add_parser('gen-task', help='Generate tasks.')

    parser.add_argument('-g', '--gff', required=True, help='The path of the GFF annotation file.')
    parser.add_argument('-G', '--gene-feature', default='gene',
                        help='The feature type name of genes, '
                             'in the 3rd column of the GFF file. (DEFAULT: gene)')
    parser.add_argument('-I', '--isoform-feature', default='transcript',
                        help='The feature type name of isoforms, '
                             'in the 3rd column of the GFF file. (DEFAULT: transcript)')
    parser.add_argument('-A', '--gene-name-attr', default='ID',
                        help='The attribute that describes gene names, '
                             'extracted from the 9th column of the GFF file. (DEFAULT: ID)')
    parser.add_argument('-B', '--isoform-name-attr', default='ID',
                        help='The attribute that describes isoform names, '
                             'extracted from the 9th column of the GFF file. (DEFAULT: ID)')

    parser.add_argument('-v', '--vcf', required=True, help='The path of the VCF file.')
    parser.add_argument('-s', '--sample', required=True, help='The name of the target sample in the VCF file.')
    parser.add_argument('-p', '--ploidy', type=int, default=2, help='The ploidy of the target sample. (DEFAULT: 2)')
    parser.add_argument('-C', '--add_chrom_prefix', action='store_true',
                        help='Add "chr" prefix to chromosome names when extracting SNPs from the VCF file.')

    parser.add_argument('-o', '--out-dir', required=True, help='The path of the output directory.')


def _add_extract_region_parser(subparsers):
    parser = subparsers.add_parser('extract-region', help='Extract long continuous regions.')

    parser.add_argument('-g', '--gff', required=True, help='The path of the GFF3 annotation file.')
    parser.add_argument('-m', '--min-region-len', type=int, default=1000,
                        help='Minimum length of regions. (DEFAULT: 1000)')

    parser.add_argument('-o', '--out', required=True, help='The path of the output file.')


def _add_stat_insert_size_parser(subparsers):
    parser = subparsers.add_parser('stat-insert-size', help='Stat insert size of paired-end BAM files.')

    parser.add_argument('-r', '--region', required=True,
                        help='The path of the file containing long continuous regions.')
    parser.add_argument('-b', '--bams', nargs='+', required=True, help='The paths of the BAMs.')

    parser.add_argument('-o', '--out-dir', required=True, help='The path of output directory.')


def _add_inference_parser(subparsers):
    parser = subparsers.add_parser('inference', help='Perform Bayesian inference on tasks.')

    parser.add_argument('-t', '--task', required=True, help='The path of task directory.')

    parser.add_argument('-b', '--bam', nargs='+', required=True, help='The paths of BAM files.')
    parser.add_argument('-L', '--read-len', type=int, required=True, help='Read length.')

    parser.add_argument('-P', '--pe', action='store_true', help='Specify the BAM files are paired-end.')
    parser.add_argument('-M', '--insert-size-mean', type=float, help='The mean of insert size.')
    parser.add_argument('-S', '--insert-size-std', type=float, help='The standard deviation of insert size.')

    parser.add_argument('-o', '--out-dir', required=True, help='The path of output directory.')

    parser.add_argument('-n', '--process', type=int, default=1, help='The number of processes used for '
                                                                     'parallel computing. (DEFAULT: 1)')
    parser.add_argument('-C', '--n-mcmc', type=int, default=500, help='The number of MCMC samples. (DEFAULT: 500)')
    parser.add_argument('-T', '--tune', type=int, default=500, help='The number of tune samples. (DEFAULT: 500)')
    parser.add_argument('-c', '--count', type=int, help='The count of tasks to be inferred. '
                                                        'If not specified, inference will be done on all tasks.')


def _add_inference_resume_parser(subparsers):
    parser = subparsers.add_parser('resume', help='Resume Bayesian inference.')

    parser.add_argument('-o', '--out-dir', required=True, help='The path of previous output directory.')

    parser.add_argument('-n', '--process', type=int, default=1, help='The number of processes used for '
                                                                     'parallel computing. (DEFAULT: 1)')
    parser.add_argument('-c', '--count', type=int, help='The count of tasks to be inferred. '
                                                        'If not specified, inference will be done on all tasks.')


def _add_stats_parser(subparsers):
    parser = subparsers.add_parser('stats', help='Stats results.')

    parser.add_argument('-d', '--result-dir', required=True, help='The paths of result directory.')


def _add_trace_parser(subparsers):
    parser = subparsers.add_parser('trace', help='Retrieve MCMC traces of a task.')
    parser.add_argument('-d', '--result-dir', required=True, help='The paths of result directory.')
    parser.add_argument('-i', '--task-id', required=True, help='The ID of the task.')


def _add_plot_parser(subparsers):
    parser = subparsers.add_parser('plot', help='Plot.')

    parser.add_argument('-d', '--result-dir', required=True, help='The path of result directory.')
    parser.add_argument('-i', '--task-id', required=True, help='The ID of the task to plot.')


def _parse_args():
    parser = argparse.ArgumentParser(prog='byase')
    subparsers = parser.add_subparsers(dest='sub_command')

    _add_gen_annotation_parser(subparsers)
    _add_extract_region_parser(subparsers)
    _add_stat_insert_size_parser(subparsers)
    _add_inference_parser(subparsers)
    _add_inference_resume_parser(subparsers)
    _add_stats_parser(subparsers)
    _add_trace_parser(subparsers)
    _add_plot_parser(subparsers)

    args = parser.parse_args()

    return args


def main():
    # Hide HTSeq warnings.
    warnings.filterwarnings('ignore')

    args = vars(_parse_args())
    mc = MessageCenter()
    args['mc'] = mc

    sub_command = args['sub_command']

    try:
        if sub_command == 'gen-task':
            from .annotation import generate_annotation
            generate_annotation(args)

        elif sub_command == 'extract-region':
            from .pe_utils import extract_region
            extract_region(args)

        elif sub_command == 'stat-insert-size':
            from .pe_utils import stat_insert_size
            stat_insert_size(args)

        elif sub_command == 'inference':
            from .inference import inference
            inference(args)

        elif sub_command == 'resume':
            from .inference import inference_resume
            inference_resume(args)

        elif sub_command == 'stats':
            from .stats import stats
            stats(args)

        elif sub_command == 'trace':
            from .stats import trace
            trace(args)

        elif sub_command == 'plot':
            from .plot import plot_task
            plot_task(args)

    except Exception as e:
        mc.log_error(e)
        traceback.print_exc()



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import List, Tuple, Set, Optional
from typing import NamedTuple

import numpy as np
import HTSeq

from ..message import MessageCenter
from .task import SNP, Task


_PHRED_QUAL_THRESHOLD = 20


class BAMParam:
    """BAM parameters.

    Attributes:
        paths: The paths of BAM files.
        read_len: Read length.
        paired_end: If the BAMs are paired-end.
        insert_size_mean: Mean of insert-size in paired-end data.
        insert_size_std: Std of insert-size in paired-end data.
    """

    def __init__(self, bam_paths: List[str], read_len: int, paired_end: bool,
                 insert_size_mean: Optional[float], insert_size_std: Optional[float]):
        self.paths = bam_paths
        self.read_len = read_len
        self.paired_end = paired_end
        self.insert_size_mean = insert_size_mean
        self.insert_size_std = insert_size_std

        if self.paired_end:
            assert self.insert_size_mean is not None
            assert self.insert_size_std is not None
        else:
            assert self.insert_size_mean is None
            assert self.insert_size_std is None


class _MatchedRegion(NamedTuple):
    iv: HTSeq.GenomicInterval
    seq: Optional[str]
    qual: Optional[np.ndarray]


class _SkippedRegion(NamedTuple):
    iv: HTSeq.GenomicInterval


class AlignStat:
    """Alignment stats.

    Attributes:
        aln: The original SAM alignment.
        aln2: The other SAM alignment if the BAM file is paired-end.
        matched_ivs: Matched genomic intervals.
        compatible_isoforms: Compatible isoforms.
        compatible_alleles: Compatible alleles.
    """

    def __init__(self, aln: HTSeq.SAM_Alignment, aln2: Optional[HTSeq.SAM_Alignment],
                 matched_ivs: List[HTSeq.GenomicInterval],
                 compatible_isoforms: Set[int], compatible_alleles: Set[int]):
        self.aln = aln
        self.aln2 = aln2
        self.matched_ivs = matched_ivs
        self.compatible_isoforms = compatible_isoforms
        self.compatible_alleles = compatible_alleles

    @property
    def invalid(self):
        """If the alignment is invalid."""
        return (len(self.compatible_isoforms) == 0) or (len(self.compatible_alleles) == 0)


class TaskAlign(Task):
    """Task alignment.

        Attributes:
            region: A GenomicArrayOfSets object containing isoforms and SNPs info.
            bam_param: The parameters of BAM files.
            mc: Message center.
        """

    def __init__(self, task: Task, bam_param: BAMParam, mc: MessageCenter):
        super().__init__(task.id, task.segment, task.phased, task.ploidy, task.snps)
        self.bam_param = bam_param
        self.mc = mc

        # Construct the region.
        self.region = HTSeq.GenomicArrayOfSets('auto', stranded=False)
        for i, isoform in enumerate(self.segment.isoforms):
            for exon in isoform.exons:
                self.region[exon] += i
        for snp in self.snps:
            self.region[snp.iv] += snp

    def _validate_aln(self, aln: HTSeq.SAM_Alignment) -> bool:
        """Validate alignment"""
        if not aln.aligned:
            return False
        if aln.not_primary_alignment or aln.pcr_or_optical_duplicate or aln.failed_platform_qc:
            return False
        if (aln.iv.start < self.segment.iv.start) or (self.segment.iv.end < aln.iv.end):
            return False
        return True

    def _extract_matched_and_skipped_regions(
            self, aln: HTSeq.SAM_Alignment) -> Tuple[List[_MatchedRegion], List[_SkippedRegion]]:
        """Get matched and skipped regions."""
        matched_regions = []  # type: List[_MatchedRegion]
        skipped_regions = []  # type: List[_SkippedRegion]

        last_matched_end = None

        for co in aln.cigar:
            co = co  # type: HTSeq.CigarOperation
            # CIGAR string 'D'(deletion) should be considered as matched to keep consistency.
            if co.type in ['M', '=', 'X', 'I', 'D']:
                co_iv = co.ref_iv  # type: HTSeq.GenomicInterval

                if co.type in ['M', '=', 'X']:
                    matched_seq = aln.read_as_aligned.seq[co.query_from: co.query_to].decode()
                    matched_qual = aln.read_as_aligned.qual[co.query_from: co.query_to]
                    matched_region = _MatchedRegion(iv=co_iv, seq=matched_seq, qual=matched_qual)
                    matched_regions.append(matched_region)

                if (last_matched_end is not None) and (last_matched_end < co_iv.start):
                    skipped_iv = HTSeq.GenomicInterval(self.segment.iv.chrom, last_matched_end, co_iv.start, '.')
                    skipped_region = _SkippedRegion(iv=skipped_iv)
                    skipped_regions.append(skipped_region)

                last_matched_end = co_iv.end
        return matched_regions, skipped_regions

    def _calc_compatible_isoforms(self, matched_regions: List[_MatchedRegion],
                                  skipped_regions: List[_SkippedRegion]) -> Set[int]:
        """Calculate compatible isoforms."""

        # Calculate compatible isoforms from matched regions.
        compatible_set = None

        for matched_region in matched_regions:
            for iv, value in self.region[matched_region.iv].steps():
                isoforms_set = set([v for v in value if isinstance(v, int)])
                if compatible_set is None:
                    compatible_set = isoforms_set
                else:
                    compatible_set &= isoforms_set

        if compatible_set is None:
            return set()

        # Calculate incompatible isoforms from skipped regions.
        incompatible_set = set()

        for skipped_region in skipped_regions:
            for iv, value in self.region[skipped_region.iv].steps():
                isoforms_set = set([v for v in value if isinstance(v, int)])
                incompatible_set |= isoforms_set

        # Incorporate compatible isoforms with incompatible isoforms.
        compatible_set -= incompatible_set
        return compatible_set

    def _calc_compatible_alleles(self, matched_regions: List[_MatchedRegion]) -> Set[int]:
        """Calculate compatible alleles."""
        compatible_set = set([i for i in range(self.ploidy)])

        for matched_region in matched_regions:
            for iv, value in self.region[matched_region.iv].steps():
                snps = [v for v in value if isinstance(v, SNP)]
                for snp in snps:
                    snp = snp  # type: SNP
                    nt_idx = snp.pos - matched_region.iv.start
                    nt = matched_region.seq[nt_idx]

                    # Filter base quality.
                    nt_qual = matched_region.qual[nt_idx]
                    if nt_qual < _PHRED_QUAL_THRESHOLD:
                        continue

                    alleles_set = set([i for (i, snp_a) in enumerate(snp.alleles) if snp_a == nt])

                    # If nt is not any of the possible alleles in SNP, it should be considered as invalid.
                    if len(alleles_set) == 0:
                        continue

                    compatible_set &= alleles_set
        return compatible_set

    def _calc_se_align_stat(self, aln: HTSeq.SAM_Alignment) -> AlignStat:
        """Calculate stats for single-end SAM alignment."""
        matched_regions, skipped_regions = self._extract_matched_and_skipped_regions(aln)
        compatible_isoforms = self._calc_compatible_isoforms(matched_regions, skipped_regions)
        compatible_alleles = self._calc_compatible_alleles(matched_regions)
        align_stat = AlignStat(aln=aln, aln2=None,
                               matched_ivs=[region.iv for region in matched_regions],
                               compatible_isoforms=compatible_isoforms,
                               compatible_alleles=compatible_alleles)
        return align_stat

    def _calc_pe_align_stat(self, aln1: HTSeq.SAM_Alignment, aln2: HTSeq.SAM_Alignment) -> AlignStat:
        """Calculate stats for paired-end-end SAM alignments pair."""
        aln1_stat = self._calc_se_align_stat(aln1)
        aln2_stat = self._calc_se_align_stat(aln2)
        assert aln1_stat.aln2 is None
        assert aln2_stat.aln2 is None

        aln_stat = AlignStat(aln=aln1_stat.aln, aln2=aln2_stat.aln,
                             matched_ivs=aln1_stat.matched_ivs + aln2_stat.matched_ivs,
                             compatible_isoforms=aln1_stat.compatible_isoforms & aln2_stat.compatible_isoforms,
                             compatible_alleles=aln1_stat.compatible_alleles & aln2_stat.compatible_alleles)
        return aln_stat

    def align_stats_iterator(self):
        """Iterator for (valid) alignment stats."""
        for bam_path in self.bam_param.paths:
            bam_reader = HTSeq.BAM_Reader(bam_path)
            alns = bam_reader.fetch(self.segment.iv.chrom, self.segment.iv.start, self.segment.iv.end)
            if self.bam_param.paired_end:
                for aln1, aln2 in HTSeq.pair_SAM_alignments_with_buffer(alns):
                    if (aln1 is None) or (aln2 is None):
                        continue
                    if self._validate_aln(aln1) and self._validate_aln(aln2):
                        aln_stat = self._calc_pe_align_stat(aln1, aln2)
                        if aln_stat.invalid:
                            continue
                        yield aln_stat
            else:
                for aln in alns:
                    if aln is None:
                        continue
                    if self._validate_aln(aln):
                        aln_stat = self._calc_se_align_stat(aln)
                        if aln_stat.invalid:
                            continue
                        yield aln_stat



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import sys
from datetime import datetime


ERROR = 40
WARNING = 30
INFO = 20
DEBUG = 10


_LEVEL_NAME_MAP = {
    ERROR: 'ERROR',
    WARNING: 'WARNING',
    INFO: 'INFO',
    DEBUG: 'DEBUG'
}


class MessageCenter:
    """Message center.

    Attributes:
        _level: Logging level.
        _log_path: The path of log file.
    """

    def __init__(self, level=INFO, log_path=None):
        self._level = level
        self._log_path = log_path

    def _handle_log(self, log):
        """Log handler."""
        if self._log_path is None:
            print(log, file=sys.stderr)
        else:
            with open(self._log_path, 'a') as f:
                f.write(log + '\n')

    def _log(self, level, msg):
        """Log message based on given level."""
        if level < self._level:
            return None
        time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        log = '{} | {} | {}'.format(time, _LEVEL_NAME_MAP[level], msg)
        self._handle_log(log)

    def log_error(self, msg):
        """Log error message."""
        self._log(ERROR, msg)

    def log_warning(self, msg):
        """Log warning message."""
        self._log(WARNING, msg)

    def log_info(self, msg):
        """Log info message."""
        self._log(INFO, msg)

    def log_debug(self, msg):
        """Log debug message."""
        self._log(DEBUG, msg)

    def handle_progress(self, msg, progress=None):
        """Handle progress"""
        if msg is not None:
            self.log_info(msg)
        else:
            assert progress is not None

    def handle_data(self, data):
        """Handle data."""
        pass



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os
from typing import List, NamedTuple
import struct


class DBError(Exception):
    """DB error."""
    def __init__(self, msg):
        super().__init__(msg)


class DBPathError(DBError):
    """DB path error."""
    def __init__(self, db_path, msg):
        super().__init__('[PATH: {}] {}'.format(db_path, msg))


class DBOperationError(DBError):
    """DB path error."""
    def __init__(self, db_path, msg):
        super().__init__('[DATABASE: {}] {}'.format(db_path, msg))


class Field(NamedTuple):
    name: str
    fmt: str


class ItemIdx(NamedTuple):
    id: str
    offset: int


_INT_FMT = '!q'
_INT_SIZE = struct.calcsize(_INT_FMT)


def _encode_int(value):
    """Encode int value."""
    return struct.pack(_INT_FMT, value)


def _read_int(file):
    """Read int from file."""
    buffer = file.read(_INT_SIZE)
    return struct.unpack(_INT_FMT, buffer)[0]


def _encode_field(field: Field, value):
    """Encode field value."""
    if field.fmt == 'I':
        assert isinstance(value, int)
        return _encode_int(value)

    if value is None:
        return _encode_int(0)

    if field.fmt == 'S':
        assert isinstance(value, str)
        value = value.encode()
    else:
        assert field.fmt == 'B'
        assert isinstance(value, bytes)
    return _encode_int(len(value)) + value


def _read_field(field: Field, file):
    """Read field from file."""
    if field.fmt == 'I':
        return _read_int(file)

    field_size = _read_int(file)
    if field_size == 0:
        return None

    buffer = file.read(field_size)
    if field.fmt == 'S':
        return buffer.decode()
    else:
        assert field.fmt == 'B'
        return buffer


def _encode_item(schema: List[Field], values):
    """Encode item."""
    assert len(schema) == len(values)
    code = b''
    for i, field in enumerate(schema):
        code += _encode_field(field, values[i])
    return _encode_int(len(code)) + code


def read_item(schema: List[Field], file):
    """Read item from file."""
    item_size = _read_int(file)
    start = file.tell()
    values = []
    for field in schema:
        value = _read_field(field, file)
        values.append(value)
    end = file.tell()
    assert item_size == end - start
    return values


class DB:
    """Database.

    Attributes:
        db_path: The path of database.
        read_only: If the database is read only.
        schema: The schema of database.
        idx_cache: The index cache.
    """

    _schema_info = [Field('field_name', 'S'), Field('field_fmt', 'S')]
    _idx_info = [Field('item_id', 'S'), Field('item_offset', 'I')]

    def __init__(self, db_path: str, schema: List[Field] = None, read_only: bool = True):
        self.db_path = db_path
        self.read_only = read_only

        if schema is not None:
            assert not self.read_only
            if os.path.exists(self.db_path):
                raise DBPathError(self.db_path, 'Database file already exists.')
            if os.path.exists(self.idx_path):
                raise DBPathError(self.idx_path, 'Database index file already exists.')
            self._store_schema(schema)

        self.schema = self._load_schema()

        if self.read_only:
            open_mode = 'rb'
        else:
            open_mode = 'ab+'

        self._db_file = open(self.db_path, open_mode)
        self._idx_file = open(self.idx_path, open_mode)

        if self.read_only:
            assert not self._db_file.writable()
            assert not self._idx_file.writable()
        else:
            assert self._db_file.writable()
            assert self._idx_file.writable()

        self.idx_cache = None

    def close(self):
        """Close."""
        self._db_file.close()
        self._idx_file.close()

    @property
    def idx_path(self):
        """The index path."""
        return self.db_path + '.idx'

    def _store_schema(self, schema: List[Field]):
        """Store field schema."""
        if self.read_only:
            raise DBOperationError(self.db_path, 'Database in read only mode, cannot store schema.')

        if 'id' not in [field.name for field in schema]:
            raise DBOperationError(self.db_path, 'Missing field "id" in schema.')
        for field in schema:
            if field.fmt not in ['I', 'S', 'B']:
                raise DBOperationError(self.db_path, 'Unknown fmt string "{}" in schema.'.format(field.fmt))

        schema_content = b''
        schema_size = 0
        for field in schema:
            content = _encode_item(DB._schema_info, (field.name, field.fmt))
            schema_content += content
            schema_size += len(content)
        with open(self.idx_path, 'wb') as file:
            file.write(_encode_int(schema_size))
            file.write(schema_content)

    def _load_schema(self):
        """Load field schema."""
        if not os.path.exists(self.idx_path):
            raise DBPathError(self.idx_path, 'Database index file not exists.')

        schema = []
        with open(self.idx_path, 'rb') as file:
            schema_size = _read_int(file)
            while file.tell() < _INT_SIZE + schema_size:
                name, fmt = read_item(DB._schema_info, file)
                schema.append(Field(name, fmt))
            assert file.tell() == _INT_SIZE + schema_size
        return schema

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _idx_iterator(self):
        """Index iterator."""
        f = self._idx_file
        f.seek(0, 2)
        file_size = f.tell()

        f.seek(0, 0)
        schema_size = _read_int(f)
        f.read(schema_size)
        while f.tell() < file_size:
            item = read_item(self._idx_info, f)
            item_id, item_offset = item
            yield ItemIdx(item_id, item_offset)

    def find_item_offset(self, item_id: str):
        """Find item offset."""
        if self.idx_cache is not None:
            return self.idx_cache[item_id] if item_id in self.idx_cache else None

        self.idx_cache = {}
        item_offset = None
        for idx in self._idx_iterator():
            self.idx_cache[idx.id] = idx.offset
            if item_id == idx.id:
                assert item_offset is None
                item_offset = idx.offset
        return item_offset

    def store_item(self, values):
        """Store item."""
        if self.read_only:
            raise DBOperationError(self.db_path, 'Database in read only mode, cannot store item.')

        item_id_idx = [field.name for field in self.schema].index('id')
        item_id = values[item_id_idx]

        if self.find_item_offset(item_id) is not None:
            raise DBOperationError(self.db_path, 'Item with ID "{}" already exists in database.'.format(item_id))

        self._db_file.seek(0, 2)
        item_offset = self._db_file.tell()

        # Store item.
        content = _encode_item(self.schema, values)
        self._db_file.write(content)
        # Store item index.
        content = _encode_item(self._idx_info, (item_id, item_offset))
        self._idx_file.write(content)

        # Update index cache.
        if self.idx_cache is not None:
            self.idx_cache[item_id] = item_offset

    def get_item_by_offset(self, item_offset: int):
        """Get item by offset."""
        self._db_file.seek(item_offset)
        values = read_item(self.schema, self._db_file)
        return values

    def get_item(self, item_id: str):
        """Get item by ID."""
        item_offset = self.find_item_offset(item_id)
        if item_offset is None:
            raise DBOperationError(self.db_path, 'Item with ID "{}" not exists in database.'.format(item_id))
        return self.get_item_by_offset(item_offset)

    def item_iterator(self):
        """Item iterator."""
        for idx in self._idx_iterator():
            yield self.get_item_by_offset(idx.offset)

    def get_all_item_ids(self):
        """Get all item IDs."""
        return [idx.id for idx in self._idx_iterator()]

    def preview(self, n: int):
        """Preview items."""
        print([field.name for field in self.schema])
        m = -1
        for k, item in enumerate(self.item_iterator()):
            m += 1
            if m >= n:
                break
            for i, v in enumerate(item):
                if isinstance(v, bytes):
                    item[i] = v[:min(len(v), 15)]
            print(k, item)



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import Optional, List, Dict
from enum import Enum
import os
from xml.etree.ElementTree import ElementTree

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
import HTSeq

from ..message import MessageCenter
from .task import Task
from .align import TaskAlign, BAMParam
from .result import TaskResultMeta


matplotlib.use('Agg')


class TaskPlotError(Exception):
    """Task plot error."""
    def __init__(self, task_id: str, msg):
        super().__init__('[PLOT ERROR] [TASK: {}] {}'.format(task_id, msg))


class TaskFig:
    """Task figure."""

    def __init__(self, fig_size, out_path: str):
        self.fig_size = fig_size
        self.out_path = out_path

    def __enter__(self):
        self.fig = plt.figure(figsize=self.fig_size)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fig.savefig(self.out_path)
        self.fig.clear()
        plt.close(self.fig)
        self._apply_patch(self.out_path)

    @staticmethod
    def _apply_patch(svg_path: str):
        """Apply patch to svg file generated by matplotlib to improve compatibility."""
        assert svg_path.endswith('.svg')
        tree = ElementTree(file=svg_path)
        root = tree.getroot()
        non_defs = []
        for child in root:
            if not child.tag.endswith('defs'):
                non_defs.append(child)
        for child in non_defs:
            root.remove(child)
            root.append(child)
        tree.write(svg_path)


_COLOR_BLUE = 'steelblue'
_COLOR_ORANGE = 'chocolate'
_COLOR_YELLOW = 'darkorange'
_COLOR_GREEN = 'forestgreen'
_COLOR_RED = 'indianred'
_COLOR_GREY = 'mediumpurple'
_COLOR_BROWN = '#8c564b'


def _nt_color(nt: str) -> str:
    """Nt color map."""
    if nt == 'A':
        return _COLOR_GREEN
    elif nt == 'C':
        return _COLOR_BLUE
    elif nt == 'G':
        return _COLOR_YELLOW
    elif nt == 'T':
        return _COLOR_RED
    else:
        return 'grey'


class _Coverage:
    """Coverage.

    Attributes:
        allele_num: Allele number.
        isoform_num: Isoform number.
        data: A List of GenomicArray objects that stores coverage.
        _determined: The determined status of each GenomicArray object.
        _determined_map: A dict that maps determined status to GenomicArray object.
    """

    def __init__(self, allele_num: Optional[int], isoform_num: Optional[int]):
        data = []
        determined_map = {}

        if (allele_num is None) or (isoform_num is None):
            determined = [True, False]
        else:
            # Order: (allele_determined, isoform_determined).
            determined = [(True, True), (True, False), (False, True), (False, False)]

        for determined_status in determined:
            ga = HTSeq.GenomicArray('auto', stranded=False, typecode='i')
            determined_map[determined_status] = ga
            data.append(ga)

        self.allele_num = allele_num
        self.isoform_num = isoform_num
        self.data = data
        self._determined = determined
        self._determined_map = determined_map

    def name(self, n: int):
        """The name of the nth GenomicArray object."""
        if n == len(self._determined):
            return 'Total'
        assert n < len(self._determined)
        determined_status = self._determined[n]
        if self.allele_num is None:
            return 'Isoform {}'.format(self.isoform_num + 1 if determined_status else '?')
        if self.isoform_num is None:
            return 'Allele {}'.format(self.allele_num + 1 if determined_status else '?')
        allele_determined, isoform_determined = determined_status
        allele_num = self.allele_num + 1 if allele_determined else '?'
        isoform_num = self.isoform_num + 1 if isoform_determined else '?'
        return 'A{}-I{}'.format(allele_num, isoform_num)

    def add(self, determined_status, ivs: List[HTSeq.GenomicInterval]):
        """Add counts for genomic intervals."""
        ga = self._determined_map[determined_status]
        for iv in ivs:
            ga[iv] += 1

    def calc_total(self):
        """Calculate total coverage."""
        total = HTSeq.GenomicArray('auto', stranded=False, typecode='i')
        for ga in self.data:
            for iv, v in ga.steps():
                if v != 0:
                    total[iv] += v
        self.data.append(total)

    def get_max(self) -> int:
        """Get max coverage."""
        max_coverage = 0
        assert len(self._determined) + 1 == len(self.data)
        total = self.data[-1]
        for _, v in total.steps():
            if v != 0:
                max_coverage = max(max_coverage, v)
        return max_coverage


class TaskPlotType(Enum):
    ASE = 'ase'


class TaskPlot(TaskAlign, TaskResultMeta):
    """Task plot.

    Attributes:
        trace: Result trace.
        trace_stats: Result trace stats.

        coord_offset_map: The coord offset map for display.
        cov_dict: A dict that stores coverage.

        out_dir: The output directory path.
        plot_type: Plot type.
    """

    def __init__(self, task: Task, bam_param: BAMParam,
                 trace: pd.DataFrame, trace_stats: pd.DataFrame,
                 out_dir: str, plot_type: TaskPlotType, mc: MessageCenter):
        super().__init__(task=task, bam_param=bam_param, mc=mc)

        self.trace = trace
        self.trace_stats = trace_stats
        self.out_dir = out_dir
        self.plot_type = plot_type

        self.coord_offset_map = None
        self.cov_dict = None  # type: Optional[Dict[tuple, _Coverage]]

    def _generate_coord_offset_map(self) -> HTSeq.GenomicArray:
        """Generate display coord offset map."""
        all_exons = self.segment.get_all_exons()
        total_len = sum([exon.end - exon.start for exon in all_exons])

        if len(all_exons) <= 1:
            inner_len = None
        else:
            inner_len = float(total_len) / 10.0 / float(len(all_exons) - 1)
            inner_len = max(int(inner_len), 5)

        coord_offset_map = HTSeq.GenomicArray('auto', stranded=False, typecode='i')
        offset = 0
        last_exon_end = None
        for exon in all_exons:
            if last_exon_end is None:
                offset -= exon.start
            else:
                offset += inner_len - (exon.start - last_exon_end)
            coord_offset_map[exon] = offset
            last_exon_end = exon.end
        return coord_offset_map

    def _get_coord(self, pos: int) -> int:
        """Get displayed coord of position."""
        iv = HTSeq.GenomicInterval(self.segment.iv.chrom, pos, pos + 1, '.')
        steps = [step for step in self.coord_offset_map[iv].steps()]
        assert len(steps) == 1
        _, offset = steps[0]
        return pos + offset

    def _load_coverage(self):
        """Load coverage."""
        cov_dict = {}  # type: Dict[tuple, _Coverage]
        for allele_num in list(range(self.ploidy)) + [None]:
            for iso_num in list(range(self.isoforms_count)) + [None]:
                if (allele_num is None) and (iso_num is None):
                    continue
                cov_dict[(allele_num, iso_num)] = _Coverage(allele_num=allele_num, isoform_num=iso_num)

        aln_num = -1
        for align_stat in self.align_stats_iterator():
            aln_num += 1
            if aln_num != 0 and aln_num % 1000 == 0:
                self.mc.handle_progress('{} alignments loaded...'.format(aln_num))

            isoform_determined, allele_determined = False, False
            if len(align_stat.compatible_isoforms) == 1:
                isoform_determined = True
            if len(align_stat.compatible_alleles) == 1:
                allele_determined = True

            ivs = align_stat.matched_ivs
            for allele_num in align_stat.compatible_alleles:
                cov_dict[(allele_num, None)].add(allele_determined, ivs)

            for iso_num in align_stat.compatible_isoforms:
                cov_dict[(None, iso_num)].add(isoform_determined, ivs)

            for allele_num in align_stat.compatible_alleles:
                for iso_num in align_stat.compatible_isoforms:
                    cov_dict[(allele_num, iso_num)].add((allele_determined, isoform_determined), ivs)

        for _, cov in cov_dict.items():
            cov.calc_total()

        return cov_dict

    def _coverage_path(self, allele_num: Optional[int], isoform_num: Optional[int]) -> str:
        """Plot path for coverage."""
        path = 'COV'
        if allele_num is not None:
            path += '_A{}'.format(allele_num)
        if isoform_num is not None:
            path += '_I{}'.format(isoform_num)
        path += '.svg'
        return os.path.join(self.out_dir, path)

    def _histogram_path(self, var_name: str) -> str:
        """Plot path for histogram."""
        path = 'HIST_{}.svg'.format(var_name)
        return os.path.join(self.out_dir, path)

    @property
    def _html_path(self) -> str:
        """Plot path for html."""
        return os.path.join(self.out_dir, 'plot_{}.html'.format(self.plot_type.value))

    def _plot_coverage(self, allele_num: Optional[int], isoform_num: Optional[int], y_lim: int):
        """Plot coverage."""
        out_path = self._coverage_path(allele_num=allele_num, isoform_num=isoform_num)
        with TaskFig(fig_size=(9, 3), out_path=out_path) as task_fig:

            x_lim = (self._get_coord(self.segment.iv.start) - 0.5,
                     self._get_coord(self.segment.iv.end - 1) + 0.5)

            # Plot isofrom.
            ax = task_fig.fig.add_axes([0.05, 0.05, 0.9, 0.1])
            ax.set_axis_off()

            if isoform_num is None:
                exons = self.segment.get_all_exons()
                snps = self.snps
            else:
                exons = self.segment.isoforms[isoform_num].exons
                snps = self.isoform_snps(isoform_num=isoform_num)
            exons = sorted(exons, key=lambda _iv: _iv.start)

            prev_exon_end = None
            for exon in exons:
                s = self._get_coord(exon.start)
                e = self._get_coord(exon.end - 1)
                ax.add_patch(Rectangle((s, 0), e - s, 1, color=_COLOR_GREY, linewidth=0, alpha=0.2))
                if prev_exon_end is not None:
                    ax.add_line(Line2D([prev_exon_end, s], [0.5, 0.5], linestyle=':', color='C4', alpha=0.2))
                prev_exon_end = e

            if allele_num is not None:
                for snp in snps:
                    x = self._get_coord(snp.pos)
                    ax.add_line(Line2D([x, x], [0, 1], color=_nt_color(snp.alleles[allele_num]), alpha=0.6))

            ax.set_xlim(x_lim)

            # Plot coverage.
            ax = task_fig.fig.add_axes([0.05, 0.2, 0.9, 0.7])

            if isoform_num is None:
                title = self.segment.gene_name
                colors = [_COLOR_BLUE, _COLOR_BROWN]
                alphas = [0.8, 0.35]
            elif allele_num is None:
                title = '{} - {}'.format(self.segment.gene_name, self.segment.isoforms[isoform_num].name)
                colors = [_COLOR_GREEN, _COLOR_BROWN]
                alphas = [0.5, 0.35]
            else:
                title = self.segment.isoforms[isoform_num].name
                colors = [_COLOR_BLUE, _COLOR_ORANGE, _COLOR_GREEN, _COLOR_RED]
                alphas = [0.8, 0.65, 0.5, 0.35]
            colors.append(_COLOR_GREY)
            alphas.append(0.2)

            cov = self.cov_dict[(allele_num, isoform_num)]
            for i, ga in enumerate(cov.data):
                x = []
                y = []
                for iv, v in ga.steps():
                    if v != 0:
                        s = self._get_coord(iv.start)
                        e = self._get_coord(iv.end - 1)
                        for k in [s-0.5, s-0.5, e+0.5, e+0.5]:
                            x.append(k)
                        for k in [0, v, v, 0]:
                            y.append(k)
                ax.fill_between(x, 0, y, lw=0, color=colors[i], alpha=alphas[i],
                                zorder=len(cov.data)-i, label=cov.name(i))

            ax.legend(loc='upper right', ncol=len(cov.data) + 1)
            ax.set_xlim(x_lim)
            ax.set_xticks([])
            ax.set_ylim(0, y_lim)

            ax.set_title(title, loc='left', fontsize=11)

    def _plot_histogram(self, var_name: str, title: str):
        """Plot histogram."""
        trace = self.trace[var_name]
        stats = self.trace_stats.loc[var_name, ]
        out_path = self._histogram_path(var_name)

        with TaskFig(fig_size=(3, 3), out_path=out_path) as task_fig:
            with plt.style.context('ggplot'):
                ax = task_fig.fig.add_axes([0.1, 0.25, 0.8, 0.65])

            ax.hist(trace, bins=np.arange(-2, 2, 0.01), density=True, histtype='stepfilled', color=_COLOR_BLUE)

            mean = stats['mean']
            hpd_2_5 = stats['hpd_2.5']
            hpd_97_5 = stats['hpd_97.5']

            ax.axvline(hpd_2_5, color='black', linestyle='--', alpha=0.3)
            ax.axvline(hpd_97_5, color='black', linestyle='--', alpha=0.3)
            ax.axvline(mean, color=_COLOR_RED, alpha=0.6)

            stat_text = 'Mean = {:.3f}\n95% HPD: ({:.3f}, {:.3f})'.format(mean, hpd_2_5, hpd_97_5)
            ax.set_xlabel(stat_text, labelpad=5, fontsize=11)
            ax.yaxis.set_major_locator(MaxNLocator(integer=True))

            ax.set_xticks(np.arange(0, 2, 0.2))
            ax.set_xlim(0, 1)

            ax.set_title(title, loc='left', fontsize=11)

    def _generate_html(self):
        """Generate html."""
        html = '<!DOCTYPE HTML>\n'
        html += '<html>\n'
        html += '<head>\n'
        html += _CSS
        html += '</head>\n'
        html += '<body>\n'
        html += '<table border="0" cellspacing="0" cellpadding="0">\n'

        def _add_td(_html: str, _td_class: str, _img_path: str):
            _html += '<td class="{}">'.format(_td_class)
            _html += '<img src="{}"/>'.format(os.path.basename(img_path))
            _html += '</td>\n'
            return _html

        if self.plot_type is TaskPlotType.ASE:
            for isoform_num in [None] + list(range(self.isoforms_count)):
                html += '<tr class="plot {}">\n'.format('g' if isoform_num is None else 'i{}'.format(isoform_num))
                for allele_num in range(self.ploidy):
                    var = self.get_var_expression(allele_num=allele_num, iso_num=isoform_num)
                    img_path = self._histogram_path(var)
                    td_class = 'plot hist hist_{}'.format(allele_num)
                    html = _add_td(html, td_class, img_path)

                    img_path = self._coverage_path(allele_num=allele_num, isoform_num=isoform_num)
                    td_class = 'plot cov cov_{}'.format(allele_num)
                    html = _add_td(html, td_class, img_path)

                for i, j in self.delta_iterator(self.ploidy):
                    var = self.get_var_diff_expression(allele_num1=i, allele_num2=j, iso_num=isoform_num)
                    img_path = self._histogram_path(var)
                    td_class = 'plot diff diff_{}_{}'.format(i, j)
                    html = _add_td(html, td_class, img_path)

                html += '</tr>\n'
        else:
            assert False

        html += '</table>\n'
        html += _SCRIPTS
        html += '</body>\n'
        html += '</html>\n'

        with open(self._html_path, 'w') as out:
            out.write(html)

    def plot(self) -> str:
        """Plot."""
        if os.path.exists(self._html_path):
            return self._html_path

        self.coord_offset_map = self._generate_coord_offset_map()
        self.cov_dict = self._load_coverage()

        self.mc.handle_progress('Plot histograms...')
        if self.plot_type is TaskPlotType.ASE:
            for allele_num in range(self.ploidy):
                var = self.get_var_expression(allele_num=allele_num, iso_num=None)
                title = 'Gene - Allele {}'.format(allele_num+1)
                self._plot_histogram(var, title)
                for iso_num in range(self.isoforms_count):
                    var = self.get_var_expression(allele_num=allele_num, iso_num=iso_num)
                    title = 'Isoform {} - Allele {}'.format(iso_num+1, allele_num+1)
                    self._plot_histogram(var, title)

            for i, j in self.delta_iterator(self.ploidy):
                var = self.get_var_diff_expression(i, j, iso_num=None)
                title = 'Gene - Diff (Allele {} & {})'.format(i+1, j+1)
                self._plot_histogram(var, title)
                for iso_num in range(self.isoforms_count):
                    var = self.get_var_diff_expression(i, j, iso_num=iso_num)
                    title = 'Isoform {} - Diff (Allele {} & {})'.format(iso_num+1, i+1, j+1)
                    self._plot_histogram(var, title)
        else:
            assert False

        self.mc.handle_progress('Plot coverages...')
        if self.plot_type is TaskPlotType.ASE:
            max_coverage = max([self.cov_dict[(allele_num, None)].get_max() for allele_num in range(self.ploidy)])
            for allele_num in range(self.ploidy):
                self._plot_coverage(allele_num=allele_num, isoform_num=None, y_lim=max_coverage)
        else:
            assert False

        for allele_num in range(self.ploidy):
            for iso_num in range(self.isoforms_count):
                self._plot_coverage(allele_num=allele_num, isoform_num=iso_num, y_lim=max_coverage)

        self._generate_html()
        return self._html_path


_CSS = """
<style>
    body {
        margin: 0;
    }    
    table {
        border-collapse: collapse;
    }
    img {
        vertical-align: bottom;
    }
</style>
"""

_SCRIPTS = """
<script>
    function update_plots(zoom, list) {
        var cells = document.getElementsByClassName("plot");
        for (var i = 0; i < cells.length; i++) {
            cells[i].style.display = "none";
            cells[i].getElementsByTagName("img")[0].style.width=zoom;
        }
        for (var n = 0; n < list.length; n++) {
            cells = document.getElementsByClassName(list[n]);
            for (var i = 0; i < cells.length; i++) {
                cells[i].style.display = "";
            }
        }
    }
</script>
"""



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os
from typing import List

import numpy as np
import scipy.stats
import HTSeq

from .message import MessageCenter


_INSERT_SIZE_NAME = 'insert_size.txt'
_INSERT_SIZE_STAT_NAME = 'insert_size_stat.txt'
_INSERT_SIZE_STAT_PLOT_NAME = 'insert_size_stat.png'


class PEUtilError(Exception):
    """Paired-end utils error."""
    def __init__(self, msg):
        super().__init__(msg)


class PEUtilPathError(PEUtilError):
    """Path error."""
    def __init__(self, path, msg):
        super().__init__('[PATH: {}] {}'.format(path, msg))


class PEUtilParseError(PEUtilError):
    """Parsing error."""
    def __init__(self, path, msg):
        super().__init__('[PARSE: {}] {}'.format(path, msg))


def _extract_long_continuous_regions(gff_path: str, min_region_len: int, out_path: str, mc: MessageCenter):
    """Extract long continuous regions with length of at least min_region_len."""
    mc.log_debug('gff_path: {}'.format(gff_path))
    mc.log_debug('min_region_len: {}'.format(min_region_len))
    mc.log_debug('out_path: {}'.format(out_path))

    mc.handle_progress('Calculating long continuous regions...')

    region = HTSeq.GenomicArray('auto', stranded=False, typecode='i')

    if not os.path.exists(gff_path):
        raise PEUtilPathError(gff_path, 'File not exists.')

    gff = HTSeq.GFF_Reader(gff_path)
    n = -1
    for ft in gff:
        n += 1
        if n != 0 and n % 100000 == 0:
            mc.handle_progress('{} lines read from GFF file...'.format(n))
        if ft.type == 'exon':
            region[ft.iv] += 1

    with open(out_path, 'w') as o:
        for iv, v in region.steps():
            if v != 0:
                region_len = iv.end - iv.start
                if region_len >= min_region_len:
                    o.write('{0}\t{1}\t{2}\t{3}\n'.format(iv.chrom, iv.start, iv.end, region_len))


def _calc_insert_size(regions_path: str, bam_paths: List[str], out_path: str, mc: MessageCenter):
    """Calculate insert sizes of each read pairs mapped to long continuous regions."""
    mc.log_debug('regions_path: {}'.format(regions_path))
    mc.log_debug('bam_paths: {}'.format(', '.join(bam_paths)))
    mc.log_debug('out_path: {}'.format(out_path))

    mc.handle_progress('Collecting insert sizes...')

    if not os.path.exists(regions_path):
        raise PEUtilPathError(regions_path, 'File not exists.')

    bam_readers = []
    for bam_path in bam_paths:
        if not os.path.exists(bam_path):
            raise PEUtilPathError(bam_path, 'File not exists.')
        bam_reader = HTSeq.BAM_Reader(bam_path)
        bam_readers.append(bam_reader)

    with open(regions_path) as f, open(out_path, 'w') as o:
        n = -1
        for row in f:
            row = row.strip('\n')
            n += 1
            if n != 0 and n % 1000 == 0:
                mc.handle_progress('{} regions processed...'.format(n))
            cells = row.split('\t')
            try:
                chrom, start, end = cells[:3]
                start, end = int(start), int(end)
            except ValueError:
                raise PEUtilParseError(regions_path, 'Incorrect file format.')
            insert_sizes = []
            for bam_reader in bam_readers:
                alns = bam_reader.fetch(chrom, start, end)
                for aln1, aln2 in HTSeq.pair_SAM_alignments_with_buffer(alns):
                    aln1 = aln1  # type: HTSeq.SAM_Alignment
                    aln2 = aln2  # type: HTSeq.SAM_Alignment
                    if (aln1 is None) or (aln2 is None):
                        continue

                    should_skip = False
                    for aln in [aln1, aln2]:
                        assert aln.aligned
                        assert aln.iv.start < aln.iv.end
                        if aln.not_primary_alignment or aln.pcr_or_optical_duplicate or aln.failed_platform_qc:
                            should_skip = True
                            break
                        if (aln.iv.start < start) or (end < aln.iv.end):
                            should_skip = True
                            break
                    if should_skip:
                        continue

                    insert_sizes.append(str(np.abs(aln1.inferred_insert_size)))
            o.write('{0}:{1}-{2}({3})\t{4}\n'.format(chrom, start, end, end - start, ';'.join(insert_sizes)))


def _stat_insert_size(insert_size_path: str, out_stat_path: str, out_plot_path: str, mc: MessageCenter):
    """Stat insert sizes."""
    mc.log_debug('insert_size_path: {}'.format(insert_size_path))
    mc.log_debug('out_stat_path: {}'.format(out_stat_path))
    mc.log_debug('out_plot_path: {}'.format(out_plot_path))

    if not os.path.exists(insert_size_path):
        raise PEUtilPathError(insert_size_path, 'File not exists.')

    mc.handle_progress('Calculating insert size stat...')

    insert_sizes = []
    with open(insert_size_path) as f:
        for row in f:
            lens = row.strip('\n').split('\t')[-1]
            if lens == '':
                continue
            lens = lens.split(';')
            for i in lens:
                insert_sizes.append(int(i))
    insert_sizes = np.array(insert_sizes)

    # Remove outliers.
    m = np.mean(insert_sizes)
    insert_sizes = insert_sizes[insert_sizes < 2 * m]

    m = np.mean(insert_sizes)
    s = np.std(insert_sizes)
    d = s / np.sqrt(m)

    with open(out_stat_path, 'w') as o:
        o.write('mean={}\nstd={}\ndispersion={}\npairs_count={}\n'.format(m, s, d, len(insert_sizes)))

    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(111)

    ax.hist(insert_sizes, bins=np.arange(min(insert_sizes), max(insert_sizes), 1), density=True,
            histtype='stepfilled', alpha=0.5)

    x = np.arange(min(insert_sizes), max(insert_sizes), 0.1)
    y = scipy.stats.norm.pdf(x, loc=m, scale=s)
    ax.plot(x, y, linestyle='--')

    ax.set_title(r'Distribution of insert size: $\mu={:.2f}$, $\sigma={:.2f}$'.format(m, s), fontsize=10)
    ax.set_xlabel('Insert size')
    ax.set_ylabel('Probability density')

    fig.savefig(out_plot_path)


def extract_region(args):
    """Extract long continuous regions."""
    gff_path = args['gff']
    min_region_len = args['min_region_len']
    out = args['out']
    mc = args['mc']
    _extract_long_continuous_regions(gff_path, min_region_len, out, mc)


def stat_insert_size(args):
    regions_path = args['region']
    bam_paths = args['bams']
    out_dir = args['out_dir']

    mc = args['mc']

    if not os.path.isdir(out_dir):
        raise PEUtilPathError(out_dir, 'Invalid directory.')

    insert_size_path = '{}/{}'.format(out_dir, _INSERT_SIZE_NAME)
    insert_size_stat_path = '{}/{}'.format(out_dir, _INSERT_SIZE_STAT_NAME)
    insert_size_stat_plot_path = '{}/{}'.format(out_dir, _INSERT_SIZE_STAT_PLOT_NAME)

    _calc_insert_size(regions_path, bam_paths, insert_size_path, mc)
    _stat_insert_size(insert_size_path, insert_size_stat_path, insert_size_stat_plot_path, mc)



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import List

from ..component import Segment, SNP


class Task:
    """Task.

    Attributes:
        id: Task ID.
        segment: The target segment.
        snps: The SNPs located on the segment.
        ploidy: The ploidy of the SNP.
        phased: If the task is phased.
    """

    def __init__(self, task_id: str, segment: Segment, phased: bool, ploidy: int, snps: List[SNP]):
        self.id = task_id
        self.segment = segment
        self.phased = phased
        self.ploidy = ploidy
        self.snps = snps

        if not self.phased:
            assert len(self.snps) == 1

        for snp in self.snps:
            assert self.phased == snp.phased
            assert self.ploidy == snp.ploidy

    @property
    def isoforms_count(self):
        """Isoforms count."""
        return len(self.segment.isoforms)

    def isoform_snps(self, isoform_num: int) -> List[SNP]:
        """SNPs located on isoform."""
        isoform = self.segment.isoforms[isoform_num]
        snps = []
        for snp in self.snps:
            for exon in isoform.exons:
                if exon.start <= snp.pos <= exon.end:
                    snps.append(snp)
                    break
        return snps



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os

from .annotation import AnnotationDB
from .inference import InferenceTool
from .result import ResultDB
from .task.result import TaskResult
from .task.plot import TaskPlot, TaskPlotType, TaskPlotError


_PLOT_DIR_NAME = 'plots'


class PlotError(Exception):
    """Plot error."""
    def __init__(self, msg):
        super().__init__(msg)


class PlotPathError(PlotError):
    """Plot path error."""
    def __init__(self, path: str, msg):
        super().__init__('[PLOT ERROR] [PATH: {}] {}'.format(path, msg))


def plot_task(args):
    """Plot task."""
    result_dir = args['result_dir']
    task_id = args['task_id']
    mc = args['mc']

    inference_tool = InferenceTool(result_dir, n_process=1, param=None, mc=mc)

    with AnnotationDB(inference_tool.param.anno_path) as anno_db:
        task = anno_db.get_task(task_id)

    with ResultDB(inference_tool.result_path) as result_db:
        record = result_db.load_record(task_id)

    if not record.success:
        raise TaskPlotError(task_id, 'The inference for the task wss failed.')

    task_result = TaskResult(task, record.trace)

    plot_dir = os.path.join(result_dir, _PLOT_DIR_NAME)
    task_plot_dir = os.path.join(plot_dir, task_id)
    for dir_path in [plot_dir, task_plot_dir]:
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)

    plot_type = TaskPlotType.ASE

    task_plot = TaskPlot(task, inference_tool.bam_param,
                         task_result.trace, record.trace_stats,
                         task_plot_dir, plot_type, mc)
    html_path = task_plot.plot()
    return html_path



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import Tuple

import numpy as np
import pandas as pd
import scipy.stats
import HTSeq
import pymc3 as pm
from pymc3.backends.tracetab import trace_to_dataframe
import theano.tensor as tt

from ..message import MessageCenter
from .task import Task
from .align import BAMParam, TaskAlign, AlignStat


class TaskInferenceError(Exception):
    """Task inference error."""
    def __init__(self, task_id, msg):
        super().__init__(msg)
        self.task_id = task_id


class _ObservedDistribution(pm.Continuous):
    """Distribution of observed data."""

    def __init__(self, _p, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.p = _p

    def logp(self, value):
        return tt.log(tt.mul(tt.dot(value, self.p)))


class TaskInference(TaskAlign):
    """Task inference."""

    def __init__(self, task: Task, bam_param: BAMParam, mcmc_samples: int, tune_samples: int, mc: MessageCenter):
        super().__init__(task, bam_param, mc)
        self.mcmc_samples = mcmc_samples
        self.tune_samples = tune_samples

    @property
    def isoform_lens(self):
        """An ndarray containing the length of each isoform."""
        return np.array([iso.length for iso in self.segment.isoforms])

    def _calc_compatible_factor(self, align_stat: AlignStat) -> np.ndarray:
        """"Calculate compatible factor."""

        # Based on compatible alleles.
        tmp = np.repeat([0], self.ploidy)
        for i in align_stat.compatible_alleles:
            tmp[i] = 1
        compatible_factor = np.repeat(tmp, self.isoforms_count)

        # Based on compatible isoforms.
        tmp = np.repeat([0], self.isoforms_count)
        for i in align_stat.compatible_isoforms:
            tmp[i] = 1
        tmp = np.tile(tmp, self.ploidy)
        compatible_factor &= tmp

        return compatible_factor

    def _transform_se_observed_data(self, aln_stat: AlignStat) -> np.ndarray:
        """Transform single-end alignment to observed data."""
        assert aln_stat.aln2 is None
        cf = self._calc_compatible_factor(aln_stat)

        mappable_reads_count = self.isoform_lens - self.bam_param.read_len + 1
        c = mappable_reads_count.astype(float)
        c[c <= 0] = np.inf
        c = 1 / c

        data = cf * np.tile(c, self.ploidy)
        return data

    def _transform_pe_observed_data(self, aln_stat: AlignStat) -> np.ndarray:
        """Transform paired-end alignments pair to observed data."""
        cf = self._calc_compatible_factor(aln_stat)

        aln1, aln2 = aln_stat.aln, aln_stat.aln2
        start = min(aln1.iv.start, aln2.iv.start)
        end = max(aln1.iv.end, aln2.iv.end)
        fragment_iv = HTSeq.GenomicInterval(self.segment.iv.chrom, start, end, '.')

        # Inferred insert sizes for each isoform.
        inferred_insert_sizes = np.repeat([0], self.isoforms_count)
        for iv, value in self.region[fragment_iv].steps():
            isoform_nums = [v for v in value if isinstance(v, int)]
            for i in isoform_nums:
                inferred_insert_sizes[i] += (iv.end - iv.start)

        iis = inferred_insert_sizes.astype(float)
        iis[iis == 0] = -np.inf

        # Mappable fragments count.
        c = self.isoform_lens.astype(float) - iis + 1
        c[c <= 0] = np.inf
        p = scipy.stats.norm.pdf(iis, self.bam_param.insert_size_mean, self.bam_param.insert_size_std)
        c = 1 / c * p

        data = cf * np.tile(c, self.ploidy)
        return data

    def _extract_observed_data(self):
        """Extract observed data."""
        data = []
        for aln_stat in self.align_stats_iterator():
            if self.bam_param.paired_end:
                d = self._transform_pe_observed_data(aln_stat)
            else:
                d = self._transform_se_observed_data(aln_stat)

            assert len(d[d < 0]) == 0
            if len(d[d > 0]) == 0:
                continue

            data.append(d)

        if len(data) == 0:
            raise TaskInferenceError(self.id, 'No observed data found.')

        return np.array(data)

    def run(self) -> Tuple[int, pd.DataFrame]:
        """Bayesian inference."""
        observed_data = self._extract_observed_data()
        self.mc.log_debug('{} reads (read pairs) are used for Bayesian inference.'.format(len(observed_data)))

        p_count = self.isoforms_count * self.ploidy

        with pm.Model():
            p = pm.Dirichlet(name='p', a=tt.stack([1 for _ in range(p_count)]), shape=p_count)
            c = np.tile(self.isoform_lens, self.ploidy)
            p_rescaled = pm.Deterministic(name='p_rescaled', var=(p * c) / tt.dot(p, c))
            _ObservedDistribution(name='observed', _p=p_rescaled, observed=observed_data)

            # Inference.
            trace = pm.sample(self.mcmc_samples, tune=self.tune_samples, chains=1, progressbar=False)

        # Convert trace to data frame.
        trace = trace_to_dataframe(trace)

        # Remove rescaled variables.
        cols = [col for col in trace.columns if not col.startswith('p_rescaled')]
        assert len(cols) == p_count
        trace = trace[cols]

        # Rename trace columns.
        rename_map = {}
        for i in range(self.isoforms_count):
            for j in range(self.ploidy):
                n = i + j * self.isoforms_count
                rename_map['p__{}'.format(n)] = 'RAW_I{}_A{}'.format(i, j)
        trace = trace.rename(columns=rename_map)

        return len(observed_data), trace



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os
from typing import List

import pandas as pd

from .message import MessageCenter
from .inference import InferenceTool
from .annotation import AnnotationDB
from .result import ResultDB, ResultRecord
from .task.result import TaskResult, TaskResultMeta


_STATS_DIR_NAME = 'stats'
_TRACE_DIR_NAME = 'trace'
_HEAD_MEAN = 'Mean'
_HEAD_HPD_WIDTH = '95% HPD Width'


class StatsError(Exception):
    """Stats error."""
    def __init__(self, msg):
        super().__init__(msg)


class StatsTool(TaskResultMeta):
    """Stats tool.

    Attributes:
        anno_path: The path of annotation.
        result_path: The path of result.
        out_dir: The path of output directory.

        mc: Message center.
    """

    def __init__(self, anno_path: str, result_path: str, out_dir: str,  mc: MessageCenter):
        self.anno_path = anno_path
        self.result_path = result_path
        self.out_dir = out_dir
        self.mc = mc

    @property
    def _ase_gene_level_stats_path(self):
        """Gene level ASE stats file path."""
        return os.path.join(self.out_dir, 'ASE_geneLevel.csv')

    @property
    def _ase_isoform_level_stats_path(self):
        """Isoform level ASE stats file path."""
        return os.path.join(self.out_dir, 'ASE_isoformLevel.csv')

    @staticmethod
    def _store(df: pd.DataFrame, file_path: str):
        """Store data frame."""
        df.to_csv(file_path, index=False)

    def _stats(self):
        """Generate stats."""
        gene_cols = ['Task ID', 'Gene ID', 'Gene Name', 'Location', 'Isoform Count', 'SNP Count']
        iso_cols = ['Task ID', 'Isoform ID', 'Gene ID', 'Gene Name', 'Isoform Number', 'Isoform Name', 'Location',
                    'SNP Count']

        gene_d = [[] for _ in gene_cols]
        iso_d = [[] for _ in iso_cols]

        def _add_mean(_data: list, _record: ResultRecord, _var_name: str):
            _var_stats = _record.trace_stats.loc[_var_name]
            _data.append(_var_stats['mean'])

        def _add_mean_and_hpd_width(_data: list, _record: ResultRecord, _var_name: str):
            _var_stats = _record.trace_stats.loc[_var_name]
            _data.append(_var_stats['mean'])
            _data.append(_var_stats['hpd_97.5'] - _var_stats['hpd_2.5'])

        def _add_row(_d: List[list], _cols: list, _row: list):
            assert len(_d) == len(_cols) == len(_row)
            for _i, _item in enumerate(_row):
                _d[_i].append(_item)

        n = -1
        ploidy = None
        with AnnotationDB(self.anno_path) as anno_db, ResultDB(self.result_path) as result_db:
            for task in anno_db.tasks_iterator():
                record = result_db.get_record(task.id)
                if not record.success:
                    continue

                if ploidy is None:
                    ploidy = task.ploidy
                    for i in range(ploidy):
                        for cols, d in [(gene_cols, gene_d),
                                        (iso_cols, iso_d)]:
                            cols.append('Allele {} Expression Mean'.format(i+1))
                            d.append([])
                    for i, j in self.delta_iterator(ploidy):
                        for cols, d in [(gene_cols, gene_d),
                                        (iso_cols, iso_d)]:
                            for measure in [_HEAD_MEAN, _HEAD_HPD_WIDTH]:
                                cols.append('Difference (Allele {} & {}) {}'.format(i+1, j+1, measure))
                                d.append([])
                else:
                    assert ploidy == task.ploidy

                assert ploidy == task.ploidy

                n += 1
                if n != 0 and n % 100 == 0:
                    self.mc.handle_progress('Stats: {} tasks processed...'.format(n))

                seg = task.segment
                location = '{}:{}-{}'.format(seg.iv.chrom, seg.iv.start, seg.iv.end)
                gene_row = [task.id, seg.id, seg.gene_name, location, seg.isoforms_count, len(task.snps)]
                for i in range(ploidy):
                    var = self.get_var_expression(allele_num=i)
                    _add_mean(gene_row, record, var)
                for i, j in self.delta_iterator(ploidy):
                    var = self.get_var_diff_expression(allele_num1=i, allele_num2=j)
                    _add_mean_and_hpd_width(gene_row, record, var)
                _add_row(gene_d, gene_cols, gene_row)

                for iso_num, iso in enumerate(seg.isoforms):
                    location = '{}:{}-{}'.format(iso.iv.chrom, iso.iv.start, iso.iv.end)
                    iso_row = [task.id, iso.id, seg.id, seg.gene_name, iso_num + 1, iso.name,
                               location, len(task.isoform_snps(iso_num))]
                    for i in range(ploidy):
                        var = self.get_var_expression(allele_num=i, iso_num=iso_num)
                        _add_mean(iso_row, record, var)
                    for i, j in self.delta_iterator(ploidy):
                        var = self.get_var_diff_expression(allele_num1=i, allele_num2=j, iso_num=iso_num)
                        _add_mean_and_hpd_width(iso_row, record, var)
                    _add_row(iso_d, iso_cols, iso_row)

        def _get_df(_d: List[list], _cols: list):
            assert len(_d) == len(_cols)
            return pd.DataFrame({_col: _d[_i] for _i, _col in enumerate(_cols)})[_cols]

        gene_df = _get_df(gene_d, gene_cols)
        iso_df = _get_df(iso_d, iso_cols)

        self._store(gene_df, self._ase_gene_level_stats_path)
        self._store(iso_df, self._ase_isoform_level_stats_path)

    def stats(self):
        """Generate stats."""
        already_exists = True
        for file_path in [self._ase_gene_level_stats_path, self._ase_isoform_level_stats_path]:
            if not os.path.exists(file_path):
                already_exists = False

        if not already_exists:
            self._stats()

        return {'gene-level path': self._ase_gene_level_stats_path,
                'isoform-level path': self._ase_isoform_level_stats_path}


def stats(args):
    """Generate stats."""
    result_dir = args['result_dir']

    mc = args['mc']  # type: MessageCenter

    mc.log_debug('result_dir: {}'.format(result_dir))

    inference_tool = InferenceTool(result_dir, n_process=1, param=None, mc=mc)

    stats_dir = os.path.join(result_dir, _STATS_DIR_NAME)
    if not os.path.exists(stats_dir):
        os.mkdir(stats_dir)

    stats_tool = StatsTool(anno_path=inference_tool.anno_path, result_path=inference_tool.result_path,
                           out_dir=stats_dir, mc=mc)
    return stats_tool.stats()


def trace(args):
    """Get MCMC traces."""
    result_dir = args['result_dir']
    task_id = args['task_id']

    mc = args['mc']  # type: MessageCenter

    mc.log_debug('result_dir: {}'.format(result_dir))
    mc.log_debug('task_id: {}'.format(task_id))

    trace_dir = os.path.join(result_dir, _TRACE_DIR_NAME)
    if not os.path.exists(trace_dir):
        os.mkdir(trace_dir)

    out_path = os.path.join(trace_dir, '{}.csv'.format(task_id))

    inference_tool = InferenceTool(result_dir, n_process=1, param=None, mc=mc)
    anno_path = inference_tool.anno_path
    result_path = inference_tool.result_path

    with AnnotationDB(anno_path) as anno_db, ResultDB(result_path) as result_db:
        task = anno_db.get_task(task_id)
        record = result_db.get_record(task_id)
        result = TaskResult(task, record.trace)
        result.trace.to_csv(out_path, index=False)



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os
from multiprocessing import Process, Queue
from typing import NamedTuple, List, Optional
import shutil

from .message import MessageCenter
from .task import BAMParam
from .annotation import AnnotationDB
from .result import ResultRecord, ResultDB


_PARAM_FILENAME = 'param.tab'
_RESULT_FILENAME = 'result.db'
_TMP_DIR_NAME = 'tmp'
_SIGNAL_END = -1


class InferenceError(Exception):
    """Inference error."""
    def __init__(self, msg):
        super().__init__(msg)


class InferencePathError(InferenceError):
    """"Path error."""
    def __init__(self, path, msg):
        super().__init__('[PATH: {}] {}'.format(path, msg))


class InferenceParseError(InferenceError):
    """"Parsing error."""
    def __init__(self, path, msg):
        super().__init__('[PARSE: {}] {}'.format(path, msg))


def _inference_task(task_id: str, anno_path: str, bam_param: BAMParam, mcmc_samples: int, tune_samples: int, q: Queue):
    """Inference for task."""
    os.environ['OMP_NUM_THREADS'] = '1'

    from .task.inference import TaskInference, TaskInferenceError
    from .task.result import TaskResult

    mc = MessageCenter()

    with AnnotationDB(anno_path) as anno_db:
        task = anno_db.get_task(task_id)

    error = None
    try:
        task_inference = TaskInference(task, bam_param, mcmc_samples, tune_samples, mc)
        fragments_count, trace = task_inference.run()
        task_result = TaskResult(task, trace)
        stats = task_result.stats()
    except Exception as e:
        fragments_count = None
        trace = None
        stats = None
        error = '{}'.format(e) if isinstance(e, TaskInferenceError) else '[UNKNOWN ERROR] {}'.format(e)
        mc.log_error('[TASK: {}] {}'.format(task_id, error))
    success = error is None

    record = ResultRecord(task_id=task_id, success=success, error_msg=error,
                          fragments_count=fragments_count, trace=trace, trace_stats=stats)

    q.put(record)


def _handle_tasks(num: int, anno_path: str, bam_param: BAMParam, mcmc_samples: int, tune_samples: int,
                  theano_cache_dir: str, in_queue: Queue, out_queue: Queue, mc: MessageCenter):
    """Handle tasks."""
    theano_cache_path = os.path.join(theano_cache_dir, str(num))
    os.environ['THEANO_FLAGS'] = 'base_compiledir={}'.format(theano_cache_path)

    q = Queue()

    n = 0
    while True:
        item = in_queue.get()
        if isinstance(item, int) and item == _SIGNAL_END:
            break
        assert isinstance(item, str)

        task_id = item
        mc.handle_progress('[PROCESS {}] Inference for task: [{}]...'.format(num, task_id))
        mc.handle_data(('process_start', task_id))

        proc = Process(target=_inference_task, args=(task_id, anno_path, bam_param, mcmc_samples, tune_samples, q))
        proc.start()
        out_queue.put(q.get())
        proc.join()

        n += 1
        if n % 100 == 0:
            if os.path.exists(theano_cache_path):
                mc.log_debug('[PROCESS {}] Clear theano compiling cache...'.format(num))
                shutil.rmtree(theano_cache_path)
    out_queue.put(_SIGNAL_END)


def _dispatch_tasks(task_ids: List[str], handlers_count: int, in_queue: Queue):
    """Dispatch tasks."""
    for task_id in task_ids:
        in_queue.put(task_id)
    for _ in range(handlers_count):
        in_queue.put(_SIGNAL_END)


def _store_result_record(result_path: str, record: ResultRecord):
    """Store result record."""
    with ResultDB(result_path, read_only=False) as result_db:
        result_db.store_record(record)


class InferenceParam(NamedTuple):
    anno_path: str
    bam_param: BAMParam
    mcmc_samples_count: int
    tune_samples_count: int


class InferenceTool:
    """Inference tool.

    Attributes:
        n_process: The count of processes.
        param: The inference param.
        out_dir: The path of output directory.
        mc: Message center.
    """

    def __init__(self, out_dir: str, n_process: int, param: Optional[InferenceParam], mc: MessageCenter):
        self.out_dir = out_dir
        self.n_process = n_process
        self.mc = mc

        if not os.path.isdir(self.out_dir):
            raise InferencePathError(out_dir, 'Invalid directory.')

        fresh_start = param is not None

        if fresh_start:
            self.param = param
            if os.path.exists(self.result_path):
                raise InferencePathError(self.result_path, 'The result file already exists.')
        else:
            self.param = self.load_param()

        # Check paths in params.
        if not os.path.isdir(self.anno_path):
            raise InferencePathError(self.anno_path, 'The annotation directory is invalid.')
        for path in self.bam_param.paths:
            if not os.path.exists(path):
                raise InferencePathError(path, 'The BAM file not exists.')

        self.mc.log_debug('n_process: {}'.format(self.n_process))
        self.mc.log_debug('mcmc_samples: {}'.format(self.param.mcmc_samples_count))
        self.mc.log_debug('tune_samples: {}'.format(self.param.tune_samples_count))
        self.mc.log_debug('out_dir: {}'.format(self.out_dir))
        self.mc.log_debug('anno_path: {}'.format(self.anno_path))
        self.mc.log_debug('bam_paths: {}'.format(self.bam_param.paths))
        self.mc.log_debug('read_len: {}'.format(self.bam_param.read_len))
        self.mc.log_debug('paired_end: {}'.format(self.bam_param.paired_end))
        self.mc.log_debug('insert_size_mean: {}'.format(self.bam_param.insert_size_mean))
        self.mc.log_debug('insert_size_std: {}'.format(self.bam_param.insert_size_std))

        if fresh_start:
            # Store params.
            self._store_param()

        if not os.path.exists(self.result_path):
            with ResultDB(self.result_path, initialize=True, read_only=False):
                pass

    @property
    def anno_path(self):
        """The path of annotation."""
        return self.param.anno_path

    @property
    def bam_param(self):
        """BAM params."""
        return self.param.bam_param

    @property
    def param_path(self):
        """The path of params."""
        return os.path.join(self.out_dir, _PARAM_FILENAME)

    @property
    def result_path(self):
        """The path of results."""
        return os.path.join(self.out_dir, _RESULT_FILENAME)

    @property
    def theano_cache_dir(self):
        """The path of Theano cache directory."""
        return os.path.join(os.path.abspath(self.out_dir), 'theano_cache')

    @property
    def tmp_dir(self):
        """The path of tmp directory."""
        return os.path.join(self.out_dir, _TMP_DIR_NAME)

    @property
    def tmp_result_path(self):
        """The path of tmp result path."""
        return os.path.join(self.tmp_dir, _RESULT_FILENAME)

    def _store_param(self):
        """Store params."""
        if os.path.exists(self.param_path):
            raise InferencePathError(self.param_path, 'The inference parameters file already exists.')

        bp = self.bam_param
        data = [('Annotation', os.path.relpath(self.anno_path, self.out_dir)),
                ('BAM', ';'.join([os.path.relpath(path, self.out_dir) for path in bp.paths])),
                ('MCMC-samples', self.param.mcmc_samples_count),
                ('Tune-samples', self.param.tune_samples_count),
                ('Read-len', bp.read_len)]
        if bp.paired_end:
            data.append(('Paired-end', '{};{}'.format(bp.insert_size_mean, bp.insert_size_std)))
        with open(self.param_path, 'w') as f:
            for k, v in data:
                f.write('{}\t{}\n'.format(k, v))

    def load_param(self):
        """Load params."""
        if not os.path.exists(self.param_path):
            raise InferencePathError(self.param_path, 'The inference parameters file not exists.')

        anno_path, bam_paths, read_len = None, None, None
        mcmc_samples_count, tune_samples_count = None, None
        paired_end, insert_size_mean, insert_size_std = False, None, None

        with open(self.param_path) as f:
            for line in f:
                line = line.strip('\n')
                k, v = line.split('\t')
                if k == 'Annotation':
                    anno_path = os.path.join(self.out_dir, v)
                elif k == 'BAM':
                    bam_paths = [os.path.join(self.out_dir, path) for path in v.split(';')]
                elif k == 'MCMC-samples':
                    mcmc_samples_count = int(v)
                elif k == 'Tune-samples':
                    tune_samples_count = int(v)
                elif k == 'Read-len':
                    read_len = int(v)
                elif k == 'Paired-end':
                    paired_end = True
                    m, n = v.split(';')
                    insert_size_mean, insert_size_std = float(m), float(n)

        if (anno_path is None) or (bam_paths is None) or (read_len is None):
            raise InferenceParseError(self.param_path, 'The file contains incomplete inference parameters.')

        bam_param = BAMParam(bam_paths=bam_paths, read_len=read_len, paired_end=paired_end,
                             insert_size_mean=insert_size_mean, insert_size_std=insert_size_std)

        return InferenceParam(anno_path=anno_path, bam_param=bam_param,
                              mcmc_samples_count=mcmc_samples_count, tune_samples_count=tune_samples_count)

    def _store_tmp_result_records(self, records_queue: Queue):
        """Store tmp result records."""
        n = 0
        finished_count = 0
        while True:
            item = records_queue.get()
            if isinstance(item, int) and item == _SIGNAL_END:
                finished_count += 1
                if finished_count == self.n_process:
                    break
                continue
            assert isinstance(item, ResultRecord)

            proc = Process(target=_store_result_record, args=(self.tmp_result_path, item))
            proc.start()
            proc.join()

            n += 1
            self.mc.handle_data(('process_end', item.task_id))
            self.mc.handle_progress('Task: [{}] finished. ({} tasks processed)'.format(item.task_id, n), progress=n)

    def _merge_results(self):
        """Merge results."""
        self.mc.handle_progress('Merging results...')

        with ResultDB(self.result_path, read_only=False) as result_db:
            result_db.merge([self.tmp_result_path])

    def _extract_task_ids(self, target_count: Optional[int]):
        """Extract task IDs to be processed."""
        with AnnotationDB(self.anno_path) as anno_db:
            all_task_ids = anno_db.get_all_task_ids()
        with ResultDB(self.result_path) as result_db:
            finished_task_ids = set(result_db.get_all_task_ids())

        n = 0
        task_ids = []
        for task_id in all_task_ids:
            if task_id not in finished_task_ids:
                task_ids.append(task_id)
                n += 1
                if (target_count is not None) and (n == target_count):
                    break
        return task_ids

    def run(self, target_count: Optional[int], target_task_ids: Optional[List[str]] = None):
        """Run inference."""
        if os.path.exists(self.tmp_dir):
            self.mc.log_warning('The last run may not be completed, the tmp directory is not removed.')
            shutil.rmtree(self.tmp_dir)
        os.mkdir(self.tmp_dir)
        with ResultDB(self.tmp_result_path, initialize=True, read_only=False):
            pass

        if target_task_ids is None:
            task_ids = self._extract_task_ids(target_count)
        else:
            task_ids = target_task_ids

        self.mc.log_debug('{} tasks will be inferred.'.format(len(task_ids)))

        in_queue = Queue()
        out_queue = Queue()

        procs = [Process(target=_dispatch_tasks, args=(task_ids, self.n_process, in_queue))]
        for i in range(self.n_process):
            procs.append(Process(target=_handle_tasks, args=(i, self.anno_path, self.bam_param,
                                                             self.param.mcmc_samples_count,
                                                             self.param.tune_samples_count,
                                                             self.theano_cache_dir,
                                                             in_queue, out_queue, self.mc)))

        for proc in procs:
            proc.start()

        self._store_tmp_result_records(out_queue)

        for proc in procs:
            proc.join()

        self._merge_results()
        self.mc.handle_progress('All tasks finished.')

        shutil.rmtree(self.tmp_dir)


def inference(args):
    """Inference."""
    anno_path = args['task']

    bam_paths = args['bam']
    read_len = args['read_len']

    paired_end = args['pe']
    insert_size_mean = args['insert_size_mean']
    insert_size_std = args['insert_size_std']

    out_dir = args['out_dir']

    n_process = args['process']

    mcmc_samples_count = 500
    tune_samples_count = 500
    if 'n_mcmc' in args:
        mcmc_samples_count = args['n_mcmc']
    if 'tune' in args:
        tune_samples_count = args['tune']

    target_count = args['count']

    mc = args['mc']

    bam_param = BAMParam(bam_paths=bam_paths, read_len=read_len, paired_end=paired_end,
                         insert_size_mean=insert_size_mean, insert_size_std=insert_size_std)
    param = InferenceParam(anno_path=anno_path, bam_param=bam_param,
                           mcmc_samples_count=mcmc_samples_count, tune_samples_count=tune_samples_count)

    inference_tool = InferenceTool(out_dir=out_dir, n_process=n_process, param=param, mc=mc)
    inference_tool.run(target_count)


def inference_resume(args):
    """Inference resume."""
    out_dir = args['out_dir']
    n_process = args['process']
    target_count = args['count']
    mc = args['mc']

    inference_tool = InferenceTool(out_dir=out_dir, n_process=n_process, param=None, mc=mc)
    inference_tool.run(target_count)



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

import os
from typing import List, Optional

import HTSeq

from .message import MessageCenter
from .component import Segment, Isoform, SNP
from .task import Task
from .db import DB, Field


class AnnotationError(Exception):
    """Annotation error."""
    def __init__(self, msg):
        super().__init__(msg)


class AnnotationPathError(AnnotationError):
    """Path error."""
    def __init__(self, path, msg):
        super().__init__('[PATH: {}] {}'.format(path, msg))


class AnnotationParseError(AnnotationError):
    """Parsing error."""
    def __init__(self, path, msg):
        super().__init__('[PARSE: {}] {}'.format(path, msg))


class AnnotationDBError(AnnotationError):
    """Database error."""
    def __init__(self, path, msg):
        super().__init__('[DATABASE: {}] {}'.format(path, msg))


def _read_gff(gff_path: str, segment_feature_name: str, isoform_feature_name: str,
              gene_name_attr: str, isoform_name_attr: str, mc: MessageCenter) -> List[Segment]:
    """Read all segments from GFF file.

    Args:
        gff_path: The path of GFF file.
        segment_feature_name: Segment feature name in GFF hierarchy.
        isoform_feature_name: Isoform feature name in GFF hierarchy.
        gene_name_attr: Name of attribute which describes gene name.
        isoform_name_attr: Name of attribute which describes isoform name.
        mc: Message center.
    """

    mc.log_debug('gff_path: {}'.format(gff_path))
    mc.log_debug('segment_feature_name: {}'.format(segment_feature_name))
    mc.log_debug('isoform_feature_name: {}'.format(isoform_feature_name))
    mc.log_debug('gene_name_attr: {}'.format(gene_name_attr))
    mc.log_debug('isoform_name_attr: {}'.format(isoform_name_attr))

    mc.handle_progress('Reading GFF file...')

    if not os.path.exists(gff_path):
        raise AnnotationPathError(gff_path, 'File not exists.')

    segments = []
    segments_dict = {}
    isoforms_dict = {}

    gff = HTSeq.GFF_Reader(gff_path)
    n = -1
    for ft in gff:
        n += 1
        if n != 0 and n % 100000 == 0:
            mc.handle_progress('{} lines read from GFF file...'.format(n))

        if ft.type == segment_feature_name:
            if gene_name_attr not in ft.attr:
                raise AnnotationParseError(gff_path, '"{}" not in attributes.'.format(gene_name_attr))
            segment = Segment(ft.attr['ID'], ft.attr[gene_name_attr], ft.iv)
            if segment.id in segments_dict:
                raise AnnotationParseError(gff_path, 'Segment ID is not unique: {}'.format(segment.id))
            segments_dict[segment.id] = segment
            segments.append(segment)
        elif ft.type == isoform_feature_name:
            if isoform_name_attr not in ft.attr:
                raise AnnotationParseError(gff_path, '"{}" not in attributes.'.format(isoform_name_attr))
            isoform = Isoform(ft.attr['ID'], ft.attr[isoform_name_attr], ft.iv)
            if isoform.id in isoforms_dict:
                raise AnnotationParseError(gff_path, 'Isoform ID is not unique: {}'.format(isoform.id))
            isoforms_dict[isoform.id] = isoform
            parent_id = ft.attr['Parent']
            if parent_id not in segments_dict:
                raise AnnotationParseError(gff_path, 'Cannot find the parent of isoform "{}", the segment feature '
                                                     'type name may be incorrect.'.format(isoform.id))
            parent = segments_dict[parent_id]
            assert parent.iv.chrom == isoform.iv.chrom
            assert parent.iv.strand == isoform.iv.strand
            parent.isoforms.append(isoform)
        elif ft.type == 'exon':
            parent_id = ft.attr['Parent']
            if parent_id not in isoforms_dict:
                raise AnnotationParseError(gff_path, 'Cannot find the parent of exon "{}", the isoform feature '
                                                     'type name may be incorrect.'.format(ft.attr['ID']))
            parent = isoforms_dict[parent_id]
            assert parent.iv.chrom == ft.iv.chrom
            assert parent.iv.strand == ft.iv.strand
            parent.exons.append(ft.iv)

    return segments


class HeteroSNPReader:
    """Heterozygous SNP reader.

    Attributes:
        vcf_path: The path of VCF file.
        sample: The name of sample to be extracted.
        ploidy: The ploidy of the sample
        add_chrom_prefix: If add "chr" prefix to chromosome.
        mc: Message center.
    """

    def __init__(self, vcf_path: str, sample: str, ploidy: int, add_chrom_prefix: bool, mc: MessageCenter):
        self.vcf_path = vcf_path
        self.sample = sample
        self.ploidy = ploidy
        self.add_chrom_prefix = add_chrom_prefix
        self.mc = mc

        if not os.path.exists(self.vcf_path):
            raise AnnotationPathError(self.vcf_path, 'File not exists.')

    def __iter__(self):
        self.mc.log_debug('vcf_path: {}'.format(self.vcf_path))
        self.mc.log_debug('sample: {}'.format(self.sample))
        self.mc.log_debug('ploidy: {}'.format(self.ploidy))
        self.mc.log_debug('add_chrom_prefix: {}'.format(self.add_chrom_prefix))

        vcf = HTSeq.VCF_Reader(self.vcf_path)
        vcf.parse_meta()

        self.mc.handle_progress('Reading VCF file...')

        n = -1
        for vc in vcf:
            n += 1
            if n != 0 and n % 500000 == 0:
                self.mc.handle_progress('{} lines read from VCF file...'.format(n))

            if self.sample not in vc.samples:
                raise AnnotationParseError(self.vcf_path, 'Sample "{}" not in VCF file.'.format(self.sample))

            gt = vc.samples[self.sample]['GT']
            if '.' in gt:
                continue
            if '/' in gt:
                phased = False
                if '|' in gt:
                    gt = gt.replace('|', '/')
                sep = '/'
            else:
                assert '|' in gt
                phased = True
                sep = '|'

            gt = gt.split(sep)
            if len(gt) != self.ploidy:
                raise AnnotationParseError(self.vcf_path,
                                           'The ploidy({}) may be inconsistent with the '
                                           'sample "{}"({}).'.format(self.ploidy, self.sample, len(gt)))

            ref_alt = [vc.ref] + vc.alt
            alleles = [ref_alt[int(g)] for g in gt]

            for allele in alleles:
                if len(allele) != 1:
                    continue
            if len(set(alleles)) < 2:
                continue

            chrom = vc.pos.chrom
            if self.add_chrom_prefix:
                chrom = 'chr{}'.format(chrom)
            pos = vc.pos.pos - 1

            snp = SNP(chrom, pos, alleles, phased)
            assert self.ploidy == snp.ploidy
            yield snp


_SEGMENT_DB_FILENAME = 'segment.db'
_ISOFORM_DB_FILENAME = 'isoform.db'
_SNP_DB_FILENAME = 'snp.db'
_TASK_DB_FILENAME = 'task.db'


class AnnotationDB:
    """Annotation Database."""

    _segment_scheme = [Field('id', 'S'), Field('gene_name', 'S'), Field('chrom', 'S'),
                       Field('start', 'I'), Field('end', 'I'), Field('strand', 'S'),
                       Field('isoforms_count', 'I'), Field('isoforms', 'S'), Field('isoforms_db_offsets', 'S')]
    _isoform_schema = [Field('id', 'S'), Field('name', 'S'), Field('start', 'I'), Field('end', 'I'),
                       Field('exons_count', 'I'), Field('exons_starts', 'S'), Field('exons_ends', 'S')]
    _snp_schema = [Field('id', 'S'), Field('chrom', 'S'), Field('pos', 'I'), Field('phased', 'I'),
                   Field('ploidy', 'I'), Field('alleles', 'S')]
    _task_schema = [Field('id', 'S'), Field('segment', 'S'), Field('phased', 'I'), Field('ploidy', 'I'),
                    Field('snps_count', 'I'), Field('snps', 'S')]

    def __init__(self, dir_path: str, initialize: bool = False, mc: Optional[MessageCenter] = None):
        self._mc = mc

        if not os.path.isdir(dir_path):
            raise AnnotationPathError(dir_path, 'Invalid directory.')

        read_only = not initialize

        self._segment_db = DB(os.path.join(dir_path, _SEGMENT_DB_FILENAME),
                              AnnotationDB._segment_scheme if initialize else None, read_only=read_only)
        self._isoform_db = DB(os.path.join(dir_path, _ISOFORM_DB_FILENAME),
                              AnnotationDB._isoform_schema if initialize else None, read_only=read_only)
        self._snp_db = DB(os.path.join(dir_path, _SNP_DB_FILENAME),
                          AnnotationDB._snp_schema if initialize else None, read_only=read_only)
        self._task_db = DB(os.path.join(dir_path, _TASK_DB_FILENAME),
                           AnnotationDB._task_schema if initialize else None, read_only=read_only)

        if not initialize:
            assert self._segment_db.read_only
            assert self._isoform_db.read_only
            assert self._snp_db.read_only
            assert self._task_db.read_only

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._segment_db.close()
        self._isoform_db.close()
        self._snp_db.close()
        self._task_db.close()

    def get_all_segment_ids(self):
        """Get all segment IDs."""
        return self._segment_db.get_all_item_ids()

    def get_segment_basic_info(self, segment_id):
        """Get segment basic information."""
        item = self._segment_db.get_item(segment_id)
        _, gene_name, chrom, start, end, _, isoforms_count, _, _ = item
        return {'gene_name': gene_name, 'location': '{}:{}-{}'.format(chrom, start, end),
                'isoforms_count': isoforms_count}

    def get_segment(self, segment_id):
        """Get segment by ID."""
        item = self._segment_db.get_item(segment_id)
        segment_id, gene_name, chrom, start, end, strand, isoforms_count, isoform_ids, isoform_offsets = item
        iv = HTSeq.GenomicInterval(chrom, start, end, strand)
        segment = Segment(segment_id, gene_name, iv)

        isoform_ids = isoform_ids.split(';')
        isoform_offsets = [int(offset) for offset in isoform_offsets.split(';')]
        assert isoforms_count == len(isoform_ids)
        assert isoforms_count == len(isoform_offsets)
        for n, isoform_offset in enumerate(isoform_offsets):
            item = self._isoform_db.get_item_by_offset(isoform_offset)
            isoform_id, isoform_name, start, end, exons_count, exons_starts, exons_ends = item
            assert isoform_id == isoform_ids[n]
            iv = HTSeq.GenomicInterval(chrom, start, end, strand)
            isoform = Isoform(isoform_id, isoform_name, iv)
            segment.isoforms.append(isoform)

            exons_starts = [int(s) for s in exons_starts.split(';')]
            exons_ends = [int(e) for e in exons_ends.split(';')]
            assert exons_count == len(exons_starts) == len(exons_ends)
            for i in range(exons_count):
                exon_iv = HTSeq.GenomicInterval(chrom, exons_starts[i], exons_ends[i], strand)
                isoform.exons.append(exon_iv)
            assert len(isoform.exons) == exons_count
        assert len(segment.isoforms) == isoforms_count
        return segment

    def get_snp(self, snp_id):
        """Get SNP by ID."""
        snp_id, chrom, pos, phased_code, ploidy, alleles = self._snp_db.get_item(snp_id)
        alleles = alleles.split(';')
        assert len(alleles) == ploidy
        if phased_code == 1:
            phased = True
        else:
            assert phased_code == 0
            phased = False
        snp = SNP(chrom, pos, alleles, phased)
        return snp

    def get_all_task_ids(self):
        """Get all task IDs."""
        return self._task_db.get_all_item_ids()

    def _construct_task(self, item):
        """Construct task from database item."""
        task_id, segment_id, phased_code, ploidy, snps_count, snp_ids = item
        if phased_code == 1:
            phased = True
        else:
            assert phased_code == 0
            phased = False
        snp_ids = snp_ids.split(';')
        assert len(snp_ids) == snps_count
        segment = self.get_segment(segment_id)
        snps = [self.get_snp(snp_id) for snp_id in snp_ids]
        task = Task(task_id, segment, phased, ploidy, snps)
        return task

    def tasks_iterator(self):
        """Tasks iterator."""
        for item in self._task_db.item_iterator():
            yield self._construct_task(item)

    def tasks_db_fields_iterator(self):
        """Tasks database fields iterator."""
        for item in self._task_db.item_iterator():
            yield {self._task_schema[n].name: item[n] for n in range(len(self._task_schema))}

    def get_task(self, task_id):
        """Get task by ID."""
        item = self._task_db.get_item(task_id)
        return self._construct_task(item)

    def store_segments(self, segments: List[Segment]):
        """Store segments into the database."""
        for segment in segments:
            for isoform in segment.isoforms:
                self._isoform_db.store_item(
                    (isoform.id, isoform.name, isoform.iv.start, isoform.iv.end, len(isoform.exons),
                     ';'.join([str(e.start) for e in isoform.exons]),
                     ';'.join([str(e.end) for e in isoform.exons])))
        for segment in segments:
            isoform_offsets = []
            for isoform in segment.isoforms:
                offset = self._isoform_db.find_item_offset(isoform.id)
                assert offset is not None
                isoform_offsets.append(str(offset))
            self._segment_db.store_item(
                (segment.id, segment.gene_name, segment.iv.chrom, segment.iv.start, segment.iv.end, segment.iv.strand,
                 len(segment.isoforms), ';'.join([i.id for i in segment.isoforms]), ';'.join(isoform_offsets)))

    def store_snps(self, snps):
        """Store SNPs into the database."""
        for snp in snps:
            snp = snp  # type: SNP
            self._snp_db.store_item(
                (snp.id, snp.iv.chrom, snp.iv.start, 1 if snp.phased else 0, len(snp.alleles), ';'.join(snp.alleles)))

    def create_tasks(self, ploidy: int, snps):
        """Create tasks."""
        vc_sites = HTSeq.GenomicArray('auto', stranded=False, typecode='O')

        for snp in snps:
            snp = snp  # type: SNP
            vc_sites[snp.iv] = snp

        self._mc.handle_progress('Creating tasks...')
        selected_snp_ids = set()
        selected_snps = []
        n = -1
        segment_ids = self.get_all_segment_ids()
        for segment_id in segment_ids:
            segment = self.get_segment(segment_id)
            phased_snp_ids, unphased_snp_ids = [], []
            for isoform in segment.isoforms:
                for exon in isoform.exons:
                    for vc_iv, snp in vc_sites[exon].steps():
                        if snp is not None:
                            if snp.id not in selected_snp_ids:
                                selected_snp_ids.add(snp.id)
                                selected_snps.append(snp)
                            if snp.phased and snp.id not in phased_snp_ids:
                                phased_snp_ids.append(snp.id)
                            if not snp.phased and snp.id not in unphased_snp_ids:
                                unphased_snp_ids.append(snp.id)
            if len(phased_snp_ids) > 0:
                n = self._create_task(n, segment.id, True, ploidy, phased_snp_ids)
            for unphased_snp_id in unphased_snp_ids:
                n = self._create_task(n, segment.id, False, ploidy, [unphased_snp_id])

        self.store_snps(selected_snps)

    def _create_task(self, n: int, segment_id: str, phased: bool, ploidy: int, snp_ids: List[str]) -> int:
        """Create task."""
        assert self._mc is not None

        n += 1
        if n != 0 and n % 10000 == 0:
            self._mc.handle_progress('{} tasks created...'.format(n))
        if phased:
            task_id = '{}-PHASED'.format(segment_id)
        else:
            assert len(snp_ids) == 1
            task_id = '{}-{}'.format(segment_id, snp_ids[0])
        self._task_db.store_item((task_id, segment_id, 1 if phased else 0, ploidy, len(snp_ids), ';'.join(snp_ids)))
        return n


def generate_annotation(args):
    """Generate annotation."""
    gff_path = args['gff']
    segment_feature_name = args['gene_feature']
    isoform_feature_name = args['isoform_feature']
    gene_name_attr = args['gene_name_attr']
    isoform_name_attr = args['isoform_name_attr']

    vcf_path = args['vcf']
    vcf_sample = args['sample']
    ploidy = args['ploidy']
    add_chrom_prefix = args['add_chrom_prefix']

    out_dir = args['out_dir']
    mc = args['mc']  # type: MessageCenter

    if not os.path.isdir(out_dir):
        raise AnnotationPathError(out_dir, 'Invalid directory.')

    with AnnotationDB(out_dir, initialize=True, mc=mc) as anno_db:
        segments = _read_gff(gff_path, segment_feature_name, isoform_feature_name,
                             gene_name_attr, isoform_name_attr, mc)
        anno_db.store_segments(segments)
        del segments
        anno_db.create_tasks(ploidy, HeteroSNPReader(vcf_path, vcf_sample, ploidy, add_chrom_prefix, mc))



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import Optional

import numpy as np
import pandas as pd

from .task import Task


class TaskResultMeta:
    """Task result meta."""

    @staticmethod
    def _get_var_name(var_type: str, iso_num: Optional[int] = None, allele_num: Optional[int] = None,
                      allele_num2: Optional[int] = None):
        """Get variable name."""
        assert (iso_num is not None) or (allele_num is not None)

        if var_type == 'expression':
            var_name = 'EXP'
        else:
            assert var_type == 'raw'
            assert (iso_num is not None) and (allele_num is not None) and (allele_num2 is None)
            var_name = 'RAW'

        if iso_num is not None:
            var_name += '_I{}'.format(iso_num)
        if allele_num is not None:
            var_name += '_A{}'.format(allele_num)
        if allele_num2 is not None:
            var_name = 'DIFF_' + var_name + '_A{}'.format(allele_num2)
        return var_name

    def _get_var_raw(self, iso_num: int, allele_num: int):
        """Get raw variable name."""
        return self._get_var_name('raw', iso_num=iso_num, allele_num=allele_num)

    def get_var_expression(self, allele_num: int, iso_num: Optional[int] = None):
        """Get expression variable name."""
        return self._get_var_name('expression', iso_num=iso_num, allele_num=allele_num)

    def get_var_diff_expression(self, allele_num1: int, allele_num2: int, iso_num: Optional[int] = None):
        """Get expression difference variable name."""
        return self._get_var_name('expression', iso_num=iso_num, allele_num=allele_num1, allele_num2=allele_num2)

    @staticmethod
    def delta_iterator(ploidy: int):
        """Iterator for delta allele number pairs."""
        for m in range(ploidy - 1):
            for n in range(m + 1, ploidy):
                yield m, n


class TaskResult(Task, TaskResultMeta):
    """Task result.

    Attributes:
        trace: Inference results.
    """

    def __init__(self, task: Task, trace: pd.DataFrame):
        super().__init__(task.id, task.segment, task.phased, task.ploidy, task.snps)
        self.trace = self.construct_full_trace(trace)

    def construct_full_trace(self, trace: pd.DataFrame) -> pd.DataFrame:
        """Construct full trace."""
        cols = trace.columns.tolist()
        trace = pd.DataFrame({col: trace[col].tolist() for col in cols}, index=trace.index)[cols]
        assert len(trace.columns) == self.isoforms_count * self.ploidy

        for n in range(self.ploidy):
            cols = [self._get_var_raw(iso_num=i, allele_num=n) for i in range(self.isoforms_count)]
            var = self.get_var_expression(allele_num=n)
            trace[var] = trace[cols].sum(axis=1)

        for i in range(self.isoforms_count):
            for n in range(self.ploidy):
                col = self._get_var_raw(allele_num=n, iso_num=i)
                all_cols = [self._get_var_raw(iso_num=i, allele_num=n) for n in range(self.ploidy)]
                total = trace[all_cols].sum(axis=1)
                var = self.get_var_expression(allele_num=n, iso_num=i)
                trace[var] = trace[col] / total

        for m, n in self.delta_iterator(self.ploidy):
            col1 = self.get_var_expression(allele_num=m)
            col2 = self.get_var_expression(allele_num=n)
            var = self.get_var_diff_expression(allele_num1=m, allele_num2=n)
            trace[var] = np.abs(trace[col1] - trace[col2])
            for i in range(self.isoforms_count):
                col1 = self.get_var_expression(allele_num=m, iso_num=i)
                col2 = self.get_var_expression(allele_num=n, iso_num=i)
                var = self.get_var_diff_expression(allele_num1=m, allele_num2=n, iso_num=i)
                trace[var] = np.abs(trace[col1] - trace[col2])

        return trace

    def stats(self) -> pd.DataFrame:
        """Stats task result."""
        import pymc3 as pm

        idxes = [col for col in self.trace.columns if not col.startswith('RAW')]

        means = []
        sds = []
        hpd_2_5s = []
        hpd_97_5s = []

        cols = ['mean', 'sd', 'hpd_2.5', 'hpd_97.5']
        data = [means, sds, hpd_2_5s, hpd_97_5s]

        for var_name in idxes:
            d_i = self.trace[var_name]
            means.append(np.mean(d_i))
            sds.append(np.std(d_i))
            hpd_2_5, hpd_97_5 = pm.hpd(d_i)
            hpd_2_5s.append(hpd_2_5)
            hpd_97_5s.append(hpd_97_5)

        df = pd.DataFrame({col: data[i] for i, col in enumerate(cols)}, index=idxes, columns=cols)
        return df



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import List

import HTSeq


class Isoform:
    """Isoform annotation.

    Attributes:
        id: Isoform ID in GFF file hierarchy.
        name: Isoform name.
        iv: the genomic location of the isoform.
        exons: A list of genomic locations, each represents an exon.
    """

    def __init__(self, isoform_id: str, isoform_name: str, iv: HTSeq.GenomicInterval):
        self.id = isoform_id
        self.name = isoform_name
        self.iv = iv
        self.exons = []  # type: List[HTSeq.GenomicInterval]

    def __str__(self):
        return '{} {} {}: {}'.format(self.id, self.name, self.iv,
                                     ','.join(['[{0},{1})'.format(e.start, e.end) for e in self.exons]))

    @property
    def length(self):
        """The total length of the isoform."""
        return sum([iv.end - iv.start for iv in self.exons])


class Segment:
    """Segment annotation.

    Attributes:
        id: Segment ID in GFF file hierarchy.
        gene_name: Gene name.
        iv: The genomic location of the segment.
        isoforms: The isoforms of the segment.
    """

    def __init__(self, segment_id: str, gene_name: str, iv: HTSeq.GenomicInterval):
        self.id = segment_id
        self.gene_name = gene_name
        self.iv = iv
        self.isoforms = []  # type: List[Isoform]

    @property
    def isoforms_count(self):
        """Isoforms count."""
        return len(self.isoforms)

    def __repr__(self):
        return '{} {} {}\n{}'.format(self.id, self.gene_name, self.iv,
                                     '\n'.join([str(i) for i in self.isoforms]))

    def get_all_exons(self) -> List[HTSeq.GenomicInterval]:
        """Get all exons."""
        exons = []
        tmp = HTSeq.GenomicArray('auto', stranded=False, typecode='i')
        for isoform in self.isoforms:
            for exon in isoform.exons:
                tmp[exon] = 1
        for iv, v in tmp.steps():
            if v == 1:
                exons.append(iv)
        return exons


class SNP:
    """SNP.

    Attributes:
        id: The ID of the SNP.
        iv: The genomic location of the SNP.
        alleles: Alleles of the SNP.
        phased: If the SNP is phased.
    """

    def __init__(self, chrom: str, pos: int, alleles: List[str], phased: bool):
        self.iv = HTSeq.GenomicInterval(chrom, pos, pos+1, '.')
        self.alleles = alleles
        self.phased = phased
        self.id = '{}-{}-{}-{}'.format(self.iv.chrom, self.iv.start,
                                       '1' if self.phased else '0', ''.join(self.alleles))

    @property
    def ploidy(self):
        """The ploidy of the SNP."""
        return len(self.alleles)

    @property
    def chrom(self):
        """The chromosome of the SNP."""
        return self.iv.chrom

    @property
    def pos(self):
        """The position of the SNP."""
        return self.iv.start

    def __repr__(self):
        return '{}:{} {}'.format(self.chrom, self.pos, ('|' if self.phased else '/').join(self.alleles))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="byase",
    version="1.0.2",
    author="Lili Dong",
    author_email="ncjllld@hit.edu.cn",
    description="A library that uses Bayesian inference to identify gene-level and isoform-level ASE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ncjllld/byase",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'byase=byase.run:main'],
    },
)



# This file is part of BYASE.
#
# BYASE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BYASE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BYASE.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Lili Dong
#

from typing import Optional, NamedTuple, List
import io

import pandas as pd

from .db import DB, Field


class ResultRecord(NamedTuple):
    """Result record."""
    task_id: str
    success: bool
    error_msg: Optional[str]
    fragments_count: Optional[int]
    trace: Optional[pd.DataFrame]
    trace_stats: Optional[pd.DataFrame]


class ResultDB:
    """Inference result database."""

    _schema = [Field('id', 'S'), Field('success', 'I'), Field('error_msg', 'S'),
               Field('fragments_count', 'I'), Field('trace', 'B'), Field('trace_stats', 'B')]

    def __init__(self, db_path: str, initialize: bool = False, read_only: bool = True):
        self.db_path = db_path

        if initialize:
            assert not read_only

        self._db = DB(db_path, schema=ResultDB._schema if initialize else None, read_only=read_only)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._db.close()

    def get_all_task_ids(self):
        """Get all task IDs."""
        return self._db.get_all_item_ids()

    @staticmethod
    def _construct_record(item):
        """Construct record from database item."""
        task_id, success, error_msg, fragments_count, trace, stats = item
        if success == 0:
            success = False
            assert error_msg is not None
            assert fragments_count == -1
            assert trace is None
            assert stats is None
            fragments_count = None
        else:
            assert success == 1
            success = True
            assert error_msg is None

            trace_buf = io.BytesIO(trace)
            trace = pd.read_parquet(trace_buf)
            trace_buf.close()
            stats_buf = io.BytesIO(stats)
            stats = pd.read_parquet(stats_buf)
            stats_buf.close()
        return ResultRecord(task_id=task_id, success=success, error_msg=error_msg,
                            fragments_count=fragments_count, trace=trace, trace_stats=stats)

    def records_iterator(self):
        """Records iterator."""
        for item in self._db.item_iterator():
            yield self._construct_record(item)

    def records_db_fields_iterator(self):
        """Records database fields iterator."""
        for item in self._db.item_iterator():
            yield {self._schema[n].name: item[n] for n in range(len(self._schema))}

    def get_record(self, task_id):
        """Get record by task ID."""
        item = self._db.get_item(task_id)
        return self._construct_record(item)

    def store_record(self, record: ResultRecord):
        """Store result record."""
        if not record.success:
            assert record.error_msg is not None
            assert record.fragments_count is None
            assert record.trace is None
            assert record.trace_stats is None
            fragments_count, trace, stats = -1, None, None
        else:
            assert record.error_msg is None

            fragments_count = record.fragments_count

            trace_buf = io.BytesIO()
            record.trace.to_parquet(trace_buf, index=False)
            trace_buf.seek(0)
            trace = trace_buf.read()
            trace_buf.close()

            stats_buf = io.BytesIO()
            record.trace_stats.to_parquet(stats_buf, index=True)
            stats_buf.seek(0)
            stats = stats_buf.read()
            stats_buf.close()
        self._db.store_item((record.task_id, 1 if record.success else 0,
                             record.error_msg, fragments_count, trace, stats))

    def load_record(self, task_id: str):
        """Load result record."""
        item = self._db.get_item(task_id)
        return self._construct_record(item)

    def merge(self, paths: List[str]):
        """Merge other results."""
        for path in paths:
            with ResultDB(db_path=path) as other:
                for record in other.records_iterator():
                    self.store_record(record)
